# cython: profile=False

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport cexp, ceq, creal, cimag, cabs
from finesse.cymath.math cimport fabs, radians, degrees
from finesse.cymath.gaussbeam cimport bp_gouy, transform_q
from finesse.cymath.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.simulations.base cimport NodeBeamParam
from finesse.frequency cimport frequency_info_t
from finesse.simulations.simulation cimport BaseSimulation
from finesse.simulations.homsolver cimport HOMSolver
from finesse.simulations.workspace cimport ABCDWorkspace
from finesse.symbols import Symbol

cimport cython

from cpython.ref cimport PyObject
from libc.stdlib cimport free, calloc


cdef extern from "math.h":
    double sin(double)
    double cos(double)


ctypedef (double*, double*, double*, double*) ptr_tuple_4


cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD


cdef class SpaceOpticalConnections:
    def __cinit__(self, HOMSolver mtx):
        # Only 1D arrays of views as spaces don't
        # couple frequencies together.
        Nf = mtx.optical_frequencies.size
        self.P1i_P2o = SubCCSView1DArray(Nf)
        self.P2i_P1o = SubCCSView1DArray(Nf)
        self.opt_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.opt_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class SpaceSignalConnections(SpaceOpticalConnections):
    def __cinit__(self, HOMSolver mtx):
        # Only 1D arrays of views as spaces don't
        # couple frequencies together.
        if not mtx.is_signal_matrix:
            raise Exception("Signal simulation not enabled")
        Nf = mtx.optical_frequencies.size
        self.SIGAMP_P1o = SubCCSView2DArray(1, Nf)
        self.SIGAMP_P2o = SubCCSView2DArray(1, Nf)
        self.SIGPHS_P1o = SubCCSView2DArray(1, Nf)
        self.SIGPHS_P2o = SubCCSView2DArray(1, Nf)
        self.H_P1o = SubCCSView2DArray(1, Nf)
        self.H_P2o = SubCCSView2DArray(1, Nf)

        self.sig_ptrs.SIGAMP_P1o = <PyObject***>self.SIGAMP_P1o.views
        self.sig_ptrs.SIGAMP_P2o = <PyObject***>self.SIGAMP_P2o.views
        self.sig_ptrs.SIGPHS_P1o = <PyObject***>self.SIGPHS_P1o.views
        self.sig_ptrs.SIGPHS_P2o = <PyObject***>self.SIGPHS_P2o.views
        self.sig_ptrs.H_P1o = <PyObject***>self.H_P1o.views
        self.sig_ptrs.H_P2o = <PyObject***>self.H_P2o.views


cdef class SpaceValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_4 ptr = (&self.L, &self.nr, &self.user_gouy_x, &self.user_gouy_y)
        cdef tuple params = ("L","nr","user_gouy_x","user_gouy_y") #
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class SpaceWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, BaseSimulation sim):
        cdef HOMSolver carrier = sim.carrier
        cdef HOMSolver signal = sim.signal

        super().__init__(
            owner, sim,
            SpaceOpticalConnections(carrier),
            SpaceSignalConnections(signal) if signal is not None else None,
            SpaceValues()
        )

        # Here we cast connections and values to a non-specific object
        # so the fill functions don't have to and cython can optimise better
        self.sco = self.carrier.connections
        self.scs = self.signal.connections if signal is not None else None
        self.sv = self.values

        self.P1i_id = sim.trace_node_index[owner.p1.i]
        self.P1o_id = sim.trace_node_index[owner.p1.o]
        self.P2i_id = sim.trace_node_index[owner.p2.i]
        self.P2o_id = sim.trace_node_index[owner.p2.o]

        self.car_p1o_idx = sim.carrier.node_id(owner.p1.o)
        self.car_p2o_idx = sim.carrier.node_id(owner.p2.o)

        if signal is not None:
            # If we have a signal simulation then we need to cache some indicies
            # for grabbing data when filling
            self.car_p1o_rhs_idx = carrier.get_node_info(owner.p1.o)["rhs_index"]
            self.car_p2o_rhs_idx = carrier.get_node_info(owner.p2.o)["rhs_index"]
            self.car_p_num_hom =   carrier.get_node_info(owner.p1.o)["nhoms"]
            self.strain_signal_enabled = owner.h.i.full_name in signal.nodes
            self.phase_signal_enabled = owner.phs.i.full_name in signal.nodes
            self.amp_signal_enabled = owner.amp.i.full_name in signal.nodes

        self.couplings = <complex_t*> calloc(sim.model_settings.num_HOMs, sizeof(complex_t))
        if not self.couplings:
            raise MemoryError()
        self.sym_abcd_B = NULL

    def __dealloc__(self):
        if self.couplings:
            free(self.couplings)

        cy_expr_free(self.sym_abcd_B)

    def compile_abcd_cy_exprs(self):
        cdef:
            object space = self.owner
            cdef object[:, ::1] M_sym = list(space._abcd_matrices.values())[0][0]

        # NOTE (sjr) Only element B of a Space ABCD matrix can possibly change
        if isinstance(M_sym[0][1], Symbol):
            ch_sym = M_sym[0][1].expand_symbols().eval(keep_changing_symbols=True)
            if isinstance(ch_sym, Symbol):
                self.sym_abcd_B = cy_expr_new()
                cy_expr_init(self.sym_abcd_B, ch_sym)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cpdef update_parameter_values(self) :
        ConnectorWorkspace.update_parameter_values(self)

        if self.sym_abcd_B != NULL:
            self.abcd[0][1] = cy_expr_eval(self.sym_abcd_B)


space_carrier_fill = FillFuncWrapper.make_from_ptr(c_space_carrier_fill)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_space_carrier_fill(ConnectorWorkspace cws) :
    r"""
    Fills the sub-matrix of the interferometer matrix held by `sim`, corresponding
    to the `space` component.

    The space component propagates a light field over a length :math:`L` with index of refraction
    :math:`n_r`. The product :math:`n_rL` is, by definition in Finesse, always a multiple of the default
    laser wavelength :math:`\lambda_0`. This defines a *macroscopic length*.

    .. _fig_space_couplings:
    .. figure:: ../images/space.*
        :align: center

        Field couplings for a space.

    The propagation only affects the phase of the field (in the plane-wave picture):

    .. math::
        s_1 = s_2 = \exp{\left(-i(\omega_0 + \Delta\omega) n_rL / c\right)}
        = \exp{\left(-i\Delta\omega n_rL / c\right)},

    where :math:`\exp{\left(-i\omega_0 n_rL / c\right)} = 1` following from the definition of macroscopic lengths
    above. The quantity :math:`\Delta\omega` is the offset to the default (angular) frequency :math:`\omega`.

    In the modal picture, Gouy phases are accumulated over space components. This results in the coupling equation
    including these phase terms such that each field of a different mode order has a different phase accumulation:

    .. math::
        s_{1, \mathrm{nm}} = s_{2, \mathrm{nm}} = \exp{\left(-i(\Delta\omega n_rL / c + n\psi_x + m\psi_y)\right)},

    where :math:`n, m` are the mode indices and :math:`\psi_x, \psi_y` are the Gouy phases of the space in tangential
    and sagittal planes, respectively. Note that if the flag `zero_tem00_gouy` is false, then :math:`n \rightarrow n + 1/2`
    and :math:`m \rightarrow m + 1/2`.

    Parameters
    ----------
    """
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        HOMSolver carrier = ws.sim.carrier
        space_optical_connections *conns = <space_optical_connections*>&ws.sco.opt_ptrs

        double pre_factor = 2 * PI * ws.sv.nr * ws.sv.L / C_LIGHT

        double gouy_x
        double gouy_y

        Py_ssize_t size
        int i
        frequency_info_t *frequencies
        frequency_info_t *freq

    # use user gouy phase if set
    gouy_x = radians(ws.sv.user_gouy_x if ws.use_user_gouy_x else ws.sv.computed_gouy_x)
    gouy_y = radians(ws.sv.user_gouy_y if ws.use_user_gouy_y else ws.sv.computed_gouy_y)

    frequencies = carrier.optical_frequencies.frequency_info
    size = carrier.optical_frequencies.size

    for i in range(size):
        freq = &frequencies[i]
        space_fill_optical_2_optical(
            conns, ws, freq, pre_factor,
            gouy_x, gouy_y,
            ws.sim.model_settings.phase_config.zero_tem00_gouy,
            ws.sim.model_settings.homs_view
        )

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void space_fill_optical_2_optical(
        space_optical_connections *conn,
        SpaceWorkspace ws,
        frequency_info_t *freq,
        double pre_factor,
        double gouy_x,
        double gouy_y,
        bint zero_tem00_gouy,
        int[:, ::1] homs_view
    ) noexcept:
    cdef:
        int n, m
        double ni

    phi = pre_factor * freq.f

    for field_idx in range(homs_view.shape[0]):
        n = homs_view[field_idx][0]
        m = homs_view[field_idx][1]

        if zero_tem00_gouy:
            ni = n
            mi = m
        else:
            ni = n + 0.5
            mi = m + 0.5

        gouy = ni * gouy_x + mi * gouy_y
        ws.couplings[field_idx] = cexp(1j * (-phi + gouy))

    # Fill diagonals with the vector
    if conn.P1i_P2o[freq.index]:
        (<SubCCSView>conn.P1i_P2o[freq.index]).fill_negative_zd_2(ws.couplings, 1)
    if conn.P2i_P1o[freq.index]:
        (<SubCCSView>conn.P2i_P1o[freq.index]).fill_negative_zd_2(ws.couplings, 1)


space_signal_fill = FillFuncWrapper.make_from_ptr(c_space_signal_fill)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_space_signal_fill(ConnectorWorkspace cws) :
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        HOMSolver signal = ws.sim.signal
        space_optical_connections *oconns = <space_optical_connections*>&ws.scs.opt_ptrs
        space_signal_connections *sconns = <space_signal_connections*>&ws.scs.sig_ptrs

        double pre_factor = 2 * PI * ws.sv.nr * ws.sv.L / C_LIGHT

        double gouy_x
        double gouy_y

        Py_ssize_t size
        int i
        frequency_info_t *frequencies

    # use user gouy phase if set
    gouy_x = radians(ws.sv.user_gouy_x if ws.use_user_gouy_x else ws.sv.computed_gouy_x)
    gouy_y = radians(ws.sv.user_gouy_y if ws.use_user_gouy_y else ws.sv.computed_gouy_y)

    frequencies = signal.optical_frequencies.frequency_info
    size = signal.optical_frequencies.size

    for i in range(size): # for each optical signal sideband
        space_fill_optical_2_optical(
            oconns, ws, &frequencies[i], pre_factor,
            gouy_x, gouy_y,
            ws.sim.model_settings.phase_config.zero_tem00_gouy,
            ws.sim.model_settings.homs_view
        )

    if ws.strain_signal_enabled:
        for i in range(size): # for each optical signal sideband
            strain_signal_fill(sconns, ws, &frequencies[i])

    if ws.phase_signal_enabled:
        for i in range(size): # for each optical signal sideband
            phase_signal_fill(sconns, ws, &frequencies[i])

    if ws.amp_signal_enabled:
        for i in range(size): # for each optical signal sideband
            amplitude_signal_fill(sconns, ws, &frequencies[i])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void phase_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
) noexcept:
    cdef:
        Py_ssize_t N = 0
        HOMSolver carrier = ws.sim.carrier
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        complex_t z = -1j * 0.5 # Just some pure phase modulation on the space outputs
        # Minus sign here because we model phase propagation as exp(-1j*(phi+A sin(Omega*t)))

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = carrier.node_field_vector_fast(ws.car_p1o_idx, freq.audio_carrier_index, &N)
    assert(c_p1_o != NULL)
    assert(ws.car_p_num_hom == N)

    c_p2_o = carrier.node_field_vector_fast(ws.car_p2o_idx, freq.audio_carrier_index, &N)
    assert(c_p2_o != NULL)
    assert(ws.car_p_num_hom == N)

    if sconns.SIGPHS_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGPHS_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.SIGPHS_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGPHS_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef inline void amplitude_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
) noexcept:
    cdef:
        Py_ssize_t N = 0
        HOMSolver carrier = ws.sim.carrier
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        # Just some pure amplitude modulate whatever
        # is coming out of the space ports
        complex_t z = 0.5

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = carrier.node_field_vector_fast(ws.car_p1o_idx, freq.audio_carrier_index, &N)
    assert(c_p1_o != NULL)
    assert(ws.car_p_num_hom == N)

    c_p2_o = carrier.node_field_vector_fast(ws.car_p2o_idx, freq.audio_carrier_index, &N)
    assert(c_p2_o != NULL)
    assert(ws.car_p_num_hom == N)

    if sconns.SIGAMP_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGAMP_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.SIGAMP_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.SIGAMP_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void strain_signal_fill(
    space_signal_connections *sconns,
    SpaceWorkspace ws,
    frequency_info_t *freq
) noexcept:
    cdef:
        Py_ssize_t N = 0
        double w_g, m_g, w_0
        const complex_t* c_p1_o = NULL
        const complex_t* c_p2_o = NULL
        complex_t z
        HOMSolver carrier = ws.sim.carrier

    # Reference for strain signal over a space...
    # Interferometer responses to gravitational waves:
    # Comparing Finesse simulations and analytical solutions
    # Charlotte Bond, Daniel Brown and Andreas Freise
    # LIGO DCC: T1300190
    # https://arxiv.org/pdf/1306.6752.pdf
    w_g = 2 * PI * ws.sim.model_settings.fsig  # GW frequency [rad/s]
    # What isn't really noted in the document is signals around multiple carrier
    # fields, but this is really just a scaling of w0
    w_0 = 2 * PI * (ws.sim.model_settings.f0 + freq.f_car[0])
    # 1/2 factor from Eq.7
    m_g = - 0.5 * w_0/w_g * sin(w_g * ws.sv.L * ws.sv.nr / 2 / C_LIGHT) # GW signal per strain [1/h] Eq.14
    phi_sb = - w_g * ws.sv.L * ws.sv.nr / 2 / C_LIGHT # Eq.15
    # audio order here takes care of minus sign options in Eq.17
    z = 1j*m_g * cexp(1j * phi_sb * freq.audio_order) # i factor from pi/2 in Eq.17
    # Carrier phase term in Eq.17 is included already using the out
    # going field at the ports below

    # Here we use the outgoing carrier field at each port.
    # This ensures that we include any gouy phase terms, etc.
    # that would otherwise need to be applied here as well,
    # compared to using the incoming fields.
    c_p1_o = carrier.node_field_vector_fast(ws.car_p1o_idx, freq.audio_carrier_index, &N)
    assert(c_p1_o != NULL)
    assert(ws.car_p_num_hom == N)

    c_p2_o = carrier.node_field_vector_fast(ws.car_p2o_idx, freq.audio_carrier_index, &N)
    assert(c_p2_o != NULL)
    assert(ws.car_p_num_hom == N)

    if sconns.H_P1o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.H_P1o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p1_o, 1 # 1D contiguous array from outview
        )

    if sconns.H_P2o[0][freq.index]:
        # Assume H only has a single frequency as input...
        (<SubCCSView>sconns.H_P2o[0][freq.index]).fill_negative_za_zd_2 (
            z, c_p2_o, 1 # 1D contiguous array from outview
        )


space_set_gouy = GouyFuncWrapper.make_from_ptr(c_set_gouy_phase)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef int c_set_gouy_phase(ABCDWorkspace cws) except -1:
    cdef:
        SpaceWorkspace ws = <SpaceWorkspace>cws
        const NodeBeamParam* q_p1o = &ws.sim.trace[ws.P1i_id]
        const NodeBeamParam* q_p2i = &ws.sim.trace[ws.P2o_id]

        double nr = ws.sv.nr

        complex_t qx_p1o_propagated
        complex_t qy_p1o_propagated

    qx_p1o_propagated = transform_q(
        ws.abcd, q_p1o.qx.q, nr, nr
    )
    qy_p1o_propagated = transform_q(
        ws.abcd, q_p1o.qy.q, nr, nr
    )

    # Can't have mismatches across spaces, these must occur at
    # components, so if a mismatch is present then report a bug
    # with an informative error message.
    if not ceq(q_p2i.qx.q, qx_p1o_propagated):
        raise RuntimeError(
            _space_mismatch_bug_report(
                ws,
                ws.P1i_id,
                ws.P2o_id,
                q_p1o.qx.q,
                q_p2i.qx.q,
                qx_p1o_propagated,
                "x"
            )
        )

    if not ceq(q_p2i.qy.q, qy_p1o_propagated):
        raise RuntimeError(
            _space_mismatch_bug_report(
                ws,
                ws.P1i_id,
                ws.P2o_id,
                q_p1o.qy.q,
                q_p2i.qy.q,
                qy_p1o_propagated,
                "y"
            )
        )

    # set workspace values
    ws.sv.computed_gouy_x = degrees(
        fabs(bp_gouy(&q_p2i.qx) - bp_gouy(&q_p1o.qx))
    )
    ws.sv.computed_gouy_y = degrees(
        fabs(bp_gouy(&q_p2i.qy) - bp_gouy(&q_p1o.qy))
    )

    return 0


cdef _space_mismatch_bug_report(
    SpaceWorkspace ws,
    Py_ssize_t node1_id,
    Py_ssize_t node2_id,
    complex_t q1,
    complex_t q2,
    complex_t q2e, # expected q
    direction, # "x" or "y"
) :
    node1_name = list(ws.sim.carrier.nodes.keys())[node1_id]
    node2_name = list(ws.sim.carrier.nodes.keys())[node2_id]

    z_diff = fabs(creal(q2e) - creal(q2))
    zr_diff = fabs(cimag(q2e) - cimag(q2))
    mag_diff = cabs(q2e - q2)

    return (
        f"Mismatch in space {ws.owner.name} "
        f"from {node1_name} -> {node2_name}:"
        f"\n    STORED q{direction} @ {node2_name} = {q2}"
        f"\n    EXPECTED q{direction} = {q2e}"
        f"\n        given q{direction} @ {node1_name} = {q1},"
        f"\n        with ABCD = {ws.abcd.base.tolist()}"
        f"\n    DIFFERENCES: |z2 - z1| = {z_diff}, |zr2 - zr1| = {zr_diff};"
        f"\n                 |q1 - q2| = {mag_diff}"
    )



"""Math functions for computing Zernike polynomial information.

TODO: write tests and document these functions properly
"""

import numpy as np
import scipy


def Rnm_p(n, m):
    """Generate radial polynomial for radial Zernikee."""
    pc = []
    m = abs(m)
    for ik in range(int((n - m) / 2) + 1):
        num = (-1) ** ik * scipy.special.factorial(int(n - ik))
        den = (
            scipy.special.factorial(int(ik))
            * scipy.special.factorial(int((n + m) / 2 - ik))
            * scipy.special.factorial(int((n - m) / 2 - ik))
        )
        pc.append(num / den)
    return np.array(pc)


def Rnm_eval(_r, _phi, n, m, a0):
    """Function to evaluate  radial components."""
    # Obtain the polynomial coeffs:
    pn = Rnm_p(n, m)
    Rnm = np.zeros(_r.shape)
    for idx, ip in enumerate(pn):
        Rnm += ip * (_r / a0) ** (n - 2 * idx)
    # Noll normalissation:
    Nnm = np.sqrt(2 * n + 2)
    return Nnm * Rnm


def ZPhi_eval(_phi, m):
    """Fuction to generate azimuthal component."""
    if m < 0:
        ZPhi = 1 / np.sqrt(np.pi) * np.sin(m * _phi)
    elif m > 0:
        ZPhi = 1 / np.sqrt(np.pi) * np.cos(m * _phi)
    else:
        ZPhi = 1 / np.sqrt(2 * np.pi) * np.ones(_phi.shape)
    return ZPhi


def Znm_eval(_r, _phi, n, m, a0):
    _Rnm = Rnm_eval(_r, _phi, n, m, a0)
    _Pnm = ZPhi_eval(_phi, m)
    return _Rnm * _Pnm


def Gen_nm(n):
    """Generate n and m  vectors containing n and m indices up to n, excluding zeroth
    mode."""
    vlen = np.sum(np.arange(2, n + 2))
    _n = np.zeros(vlen)
    _m = np.zeros(vlen)
    for iN in range(1, n + 1):
        iStart = np.sum(np.arange(2, iN - 1 + 2))
        iStop = np.sum(np.arange(2, iN + 2))
        _n[iStart:iStop] = iN
        for im in range(iN + 1):
            _m[iStart + im] = -iN + im * 2
    return _n, _m



from finesse.cymath.cmatrix cimport CCSMatrix
from ..homsolver cimport HOMSolver
from finesse.cymath cimport complex_t

cdef class SparseSolver(HOMSolver):
    """
    This class represents a sparse solver for a given system. It inherits from
    HOMSolver.

    Attributes
    ----------
    _M : CCSMatrix
        The compressed column sparse matrix representing the system
    """
    cdef:
        readonly CCSMatrix _M



"""Contains tests that ensure the map coordinate systems give the same effect as the BH
calculations, which are checked in the map coordinate file."""

import numpy as np
import finesse
from finesse.knm import Map
import pytest


def test_tilt_mirror_bh_map_tilt_cavity_scan():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 R=0.99 T=0.01 Rc=-1935
    m m2 R=1 T=0 Rc=2245
    link(l1, m1, 3994, m2)
    cav c m2.p1.o
    xaxis(m2.phi, lin, -10, 190, 1000)
    pd P m1.p2.i
    ad A00 m1.p2.i n=0 m=0 f=0
    ad A10 m1.p2.i n=1 m=0 f=0
    ad A01 m1.p2.i n=0 m=1 f=0
    modes(maxtem=2)
    """
    )

    model.beam_trace()  # initial trace to get beam sizes for maps
    qx = model.m2.p1.i.qx
    qy = model.m2.p1.i.qy
    x = np.linspace(-4 * qx.w, 4 * qx.w, 201)
    y = np.linspace(-4 * qy.w, 4 * qy.w, 200)
    X, Y = np.meshgrid(x, y)
    # postive yaw is negative X
    # postive pitch is positive Y
    disp_map = -100e-9 * X + 33e-9 * Y
    model.m2.surface_map = Map(x, y, opd=disp_map)

    out_map = model.run()

    # Switch off maps and use BH
    model.m1.surface_map = None
    model.m2.surface_map = None
    model.m2.xbeta = 100e-9
    model.m2.ybeta = 33e-9
    out_bh = model.run()

    assert abs(out_map["P"] - out_bh["P"]).max() < 1e-9
    assert abs(out_map["A00"] - out_bh["A00"]).max() < 1e-10
    assert abs(out_map["A10"] - out_bh["A10"]).max() < 1e-10
    assert abs(out_map["A01"] - out_bh["A01"]).max() < 1e-10


def test_tilt_mirror_map_reflection():
    """Tests that tilting a mirror map does the same as tilting with bayer-helms
    calculations."""
    model = finesse.Model()
    model.parse(
        """
    l l1
    m m1 R=1 T=0
    l l2
    link(l1, 100000, m1, l2)
    fd Er1 l1.p1.i 0
    fd Er2 l2.p1.i 0
    modes(maxtem=1)
    gauss g1 m1.p1.i w=10e-3 Rc=inf
    """
    )
    x = y = np.linspace(-0.05, 0.05, 1000)
    X, Y = np.meshgrid(x, y)
    # mirror surface has positive z in port 1 surface normal direction
    # pitching surface means surface moves in +z direction in +y
    # yawing surface means surface moves in -z direction in +x
    Z = -1e-6 * X + 0.3e-6 * Y

    model.m1.surface_map = Map(x, y, opd=Z)
    out_map = model.run()

    model.m1.surface_map = None
    model.m1.xbeta = 1e-6
    model.m1.ybeta = 0.3e-6
    out_bh = model.run()
    # positive pitch and yaw on reflections mean beam moves in
    # negative x and y on port 1 side
    # on port 2 side, spot should move in negative x but in positive y
    # directionality tested in BH coordinate tests.
    assert np.allclose(out_map["Er1"], out_bh["Er1"], atol=1e-12, rtol=1e-12)
    assert np.allclose(out_map["Er2"], out_bh["Er2"], atol=1e-12, rtol=1e-12)


def test_mirror_prism_tilt():
    """Tilt prism to get beam displacement.

    Beam displacement should be the same either direction for x positive prism tilt
    gives negative displacement on transmission y displacement is positive for +pitch
    from from side 1 to 2 from side 2 to 1 it gest a negative y shift.
    """
    model = finesse.Model()
    model.parse(
        """
    l l1
    m m1 R=0 T=1
    s sub m1.p2 m2.p1 nr=1.45 L=0.1
    m m2 R=0 T=1
    l l2
    link(l1, 0, m1)
    link(m2, 0, l2)

    fd Et1 l2.p1.i 0
    fd Et2 l1.p1.i 0

    modes(maxtem=1)
    gauss g1 m2.p2.o w=10e-3 Rc=inf
    """
    )
    x = y = np.linspace(-0.05, 0.05, 1000)
    X, Y = np.meshgrid(x, y)
    Z = -1e-6 * X + 0.3e-6 * Y

    model.m1.surface_map = Map(x, y, opd=Z)
    model.m2.surface_map = model.m1.surface_map

    out_map = model.run()

    model.m1.surface_map = None
    model.m2.surface_map = None
    model.m1.xbeta = 1e-6
    model.m2.xbeta = 1e-6
    model.m1.ybeta = 0.3e-6
    model.m2.ybeta = 0.3e-6

    out_bh = model.run()
    assert np.allclose(out_map["Et1"], out_bh["Et1"], atol=1e-12, rtol=1e-12)
    assert np.allclose(out_map["Et2"], out_bh["Et2"], atol=1e-12, rtol=1e-12)


@pytest.mark.parametrize("Rcx", [100, -100])
@pytest.mark.parametrize("Rcy", [100, -100])
def test_mirror_RoC_in_map(Rcx, Rcy):
    """A map of R**2/2/Rc should describe the same surface as setting the Rc parameters.

    The map must use the focusing element option so that the full beam propagation is
    done using the map scattering rather than a mix of BH and map
    """
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    m m1 R=1 T=0 Rc=[{Rcx},{Rcy}]
    l l2
    link(l1, m1.p1)
    link(l2, m1.p2)

    fd Er1 m1.p1.o 0
    fd Er2 m1.p2.o 0

    modes(maxtem=0)

    gauss g1 m1.p1.i wx=10e-3 wy=10e-3 Rcx=m1.Rcx Rcy=m1.Rcy
    """
    )
    x = y = np.linspace(-0.05, 0.05, 1000)
    X, Y = np.meshgrid(x, y)
    Z = X ** 2 / 2 / model.m1.Rcx.value + Y ** 2 / 2 / model.m1.Rcy.value

    model.m1.surface_map = Map(x, y, opd=Z, is_focusing_element=True)

    out_map = model.run()
    # Ensure all power is in HG00
    assert np.allclose(abs(out_map["Er1"][0]), 1, atol=1e-12, rtol=1e-12)
    assert np.allclose(abs(out_map["Er2"][0]), 1, atol=1e-12, rtol=1e-12)


def test_mirror_map_bh_tilt_cancel():
    """Put opposite tilt on mirror using maps and BH and it should cancel out.

    Tests Knm merging.
    """
    model = finesse.Model()
    model.parse(
        """
    l l1
    m m1 R=1 T=0
    l l2
    link(l1, m1, l2)
    fd Er1 l1.p1.i 0
    fd Er2 l2.p1.i 0
    modes(maxtem=2)
    gauss g1 m1.p1.i w=10e-3 Rc=inf
    """
    )
    x = y = np.linspace(-0.05, 0.05, 1000)
    X, Y = np.meshgrid(x, y)
    # mirror surface has positive z in port 1 surface normal direction
    # pitching surface means surface moves in +z direction in +y
    # yawing surface means surface moves in -z direction in +x
    Z = -1e-6 * X + 1e-6 * Y

    model.m1.surface_map = Map(x, y, opd=Z)
    model.m1.xbeta = -1e-6
    model.m1.ybeta = -1e-6
    out = model.run()
    # All power should be back in HG00 mode
    assert np.allclose(abs(out["Er1"][0]), 1)
    assert np.allclose(abs(out["Er2"][0]), 1)



cimport cython

# At some point we could make these all arrays and have some
# temperature dependence that could be interpolated
cdef struct material:
    double alpha   # thermal expansion coefficient
    double nr      # refractive Index
    double dndT    # dn/dt
    double kappa   # thermal conductivity
    double emiss   # emissivity
    double poisson # Poisson's ratio
    double E       # Young's modulus
    double rho     # density
    double C       # specific heat
    double T       # reference temperature

@cython.auto_pickle(True)
cdef class Material(object):
    cdef:
        material values



"""Ring heater thermal equations for cylindrical mirrors

Equations all based on :cite:`Ramette:16` and :cite:`T2500014`:

"""


from scipy.special import iv
from scipy.optimize.cython_optimize cimport brentq

import numpy as np
cimport numpy as np

cdef double SIGMA_F = 5.6710e-8

cdef extern from "math.h":
    double tan(double arg) nogil

cdef struct f_args:
    double tau
    double h_2a

# find solutions of Eq 11
cdef double f_v_m(double x, void *args) noexcept:
    cdef f_args *fargs = <f_args*>args
    return x + fargs.tau * tan(x * fargs.h_2a)

cdef double f_u_m(double x, void *args) noexcept:
    cdef f_args *fargs = <f_args*>args
    return x - fargs.tau * 1/tan(x * fargs.h_2a)

cpdef zeros_v_m(double a, double h, double tau, int m_max, double accuracy) :
    """Compute the roots of the equation

    .. math::
        v_m + \tau tan(v_m h/(2a)) = 0

    Parameters
    ----------
    a : float
        Mirror radius [m]
    h : float
        Mirror thickness [m]
    tau : float
        Reduced radiation constant
    n_max : int
        Max bessel order to compute up to
    accuracy : double
        absolute error on zero finding brentq algorithm

    Returns
    -------
    v_m : array_like
        Array of m_max zeros
    """
    cdef:
        Py_ssize_t n
        double start = 0
        double dx = 0
        f_args args

    if tau == 0:
        raise RuntimeError("tau should be > 0")

    args.tau = tau
    args.h_2a = h/(2*a)

    cdef double[::1] zeros = np.zeros((m_max, ), dtype=np.float64)
    dx = np.pi/h*a # separation between each tan asymptote
    # the tan function needs an ever increasing offset to push
    # it past the infinity point back to a -inf where it looks
    # for a root above that x
    offset = 1e-15
    for n in range(0, m_max):
        start = dx*(2*n+1)
        while f_v_m(start+offset, &args) > 0:
            offset *= 10
        zeros[n] = brentq(f_v_m, start+offset, dx*(2*n+2), <f_args *> &args, accuracy, 0, 1000000, NULL)

    return np.asarray(zeros)


cpdef zeros_u_m(double a, double h, double tau, int m_max, double accuracy) :
    """Compute the roots of the equation

    .. math::
        u_m - \tau \cot(u_m h/(2a)) = 0

    Parameters
    ----------
    a : float
        Mirror radius [m]
    h : float
        Mirror thickness [m]
    tau : float
        Reduced radiation constant
    n_max : int
        Max bessel order to compute up to
    accuracy : double
        absolute error on zero finding brentq algorithm

    Returns
    -------
    u_m : array_like
        Array of m_max zeros
    """
    cdef:
        Py_ssize_t n
        double start = 0
        double end = 0
        double dx = 0
        f_args args

    if tau == 0:
        raise RuntimeError("tau should be > 0")

    args.tau = tau
    args.h_2a = h/(2*a)

    cdef double[::1] zeros = np.zeros((m_max, ), dtype=np.float64)
    dx = np.pi/h*a # separation between each tan asymptote
    # floating point errors in tan mean there has to be slight
    # offsets applied to the start and end bounds to ensure
    # a root is bound
    start_offset = 1e-15
    end_offset = 1e-15
    for n in range(0, m_max):
        start = dx*(2*n)
        end = dx*2*(n+1)
        while f_u_m(start+start_offset, &args) >= 0:
            start_offset *= 10
        while f_u_m(end-end_offset, &args) <= 0:
            end_offset *= 10
        zeros[n] = brentq(f_u_m, start+start_offset, end-end_offset, <f_args *> &args, accuracy, 0, 1000000, NULL)

    return np.asarray(zeros)


def substrate_temperature(r, z, a, b, c, h,  material, barrel_material=None, T_ext=293.15, m_max=10, root_accuracy=1e-12):
    """Computes the substrate temperature distribution per
    Watt absorbed ring heater power.

    Parameters
    ----------
    r : array_like
        Radial vector [m]
    z : array_like
        Longitudinal vector [m]
    a : float
        Mirror radius [m]
    b : float
        Ring heater lower boundary relative to center [m]
    c : float
        Ring heater upper boundary relative to center [m]
    h : float
        Mirror thickness [m]
    material : :class:`finesse.materials.Material`
        Mirror substrate material, see :py:mod:`finesse.materials`
    barrel material: :class:`finesse.materials.Material`
        Barrel coating material, see :py:mod:`finesse.materials`
    T_ext : float, optional
        External temperature surrounding mirror
    m_max : int, optional
        Number of zeros to find, i.e. analytic order
    root_accuracy: float, optional
        Accuracy of root finding

    Returns
    -------
    T_rh_per_W : array_like
        2D Array of temperature distribution over [r, z]

    Notes
    -----
    Typo for A_m terms in manuscript, sin terms are functions of
    u_m, not v_m. See :cite:`Ramette:16`.
    """
    if barrel_material:
        emiss_face = material.emiss
        emiss_edge = barrel_material.emiss
    else:
        emiss_face = material.emiss
        emiss_edge = material.emiss

    tau_face = 4 * emiss_face * SIGMA_F * T_ext**3 * a / material.kappa
    tau_edge = 4 * emiss_edge * SIGMA_F * T_ext**3 * a / material.kappa

    v_m = zeros_v_m(a, h, tau_face, m_max, root_accuracy)
    u_m = zeros_u_m(a, h, tau_face, m_max, root_accuracy)
    factor = 2 / (material.kappa*2*np.pi*a*(c-b))

    v_m_c_a = v_m * c / a
    v_m_b_a = v_m * b / a
    u_m_c_a = u_m * c / a
    u_m_b_a = u_m * b / a

    u_m_h_a = u_m * h / a
    v_m_h_a = v_m * h / a
    u_m_z_a = u_m * z[:, np.newaxis] / a
    u_m_r_a = u_m * r[:, np.newaxis] / a
    v_m_z_a = v_m * z[:, np.newaxis] / a
    v_m_r_a = v_m * r[:, np.newaxis] / a

    # Typo in paper for A_m, numerators sin terms are functions of u_m NOT v_m
    A_m = factor * (np.sin(u_m_c_a) - np.sin(u_m_b_a))
    A_m /= (u_m_h_a + np.sin(u_m_h_a)) * (iv(1, u_m) * u_m + iv(0, u_m)*tau_edge) / a # Eq 18

    B_m = factor * (-np.cos(v_m_c_a) + np.cos(v_m_b_a))
    B_m /= (v_m_h_a - np.sin(v_m_h_a)) * (iv(1, v_m) * v_m + iv(0, v_m)*tau_edge) / a # Eq 18

    T_ss_per_W = (A_m * np.cos(u_m_z_a)[:, np.newaxis] * iv(0, u_m_r_a) # Eq 9
            + B_m * np.sin(v_m_z_a)[:, np.newaxis] * iv(0, v_m_r_a)).sum(2) # Eq 10

    return T_ss_per_W


def thermal_lens(r, a, b, c, h, material, barrel_material = None, T_ext=293.15, m_max=10, root_accuracy=1e-12):
    """Computes the substrate thermal lens per
    Watt absorbed ring heater power.

    Parameters
    ----------
    r : array_like
        Radial vector [m]
    a : float
        Mirror radius [m]
    b : float
        Ring heater lower boundary relative to center [m]
    c : float
        Ring heater upper boundary relative to center [m]
    h : float
        Mirror thickness [m]
    material : :class:`finesse.materials.Material`
        Mirror substrate material, see :py:mod:`finesse.materials`
    barrel_material : :class:`finesse.materials.Material`
        Barrel coating material, see :py:mod:`finesse.materials`
    T_ext : float, optional
        External temperature surrounding mirror
    m_max : int, optional
        Number of zeros to find, i.e. analytic order
    root_accuracy: float, optional
        Accuracy of root finding

    Returns
    -------
    Z_rh_per_W : array_like
        1D radial vector of optical path difference from
        propagating through substrate.

    Notes
    -----
    Typo for A_m terms in manuscript, sin terms are functions of
    u_m, not v_m. See :cite:`Ramette:16`.
    """
    if barrel_material:
        emiss_face = material.emiss
        emiss_edge = barrel_material.emiss
    else:
        emiss_face = material.emiss
        emiss_edge = material.emiss

    tau_face = 4 * emiss_face * SIGMA_F * T_ext**3 * a / material.kappa
    tau_edge = 4 * emiss_edge * SIGMA_F * T_ext**3 * a / material.kappa
    u_m = zeros_u_m(a, h, tau_face, m_max, root_accuracy)
    factor = 2 / (material.kappa*2*np.pi*a*(c-b))

    u_m_c_a = u_m * c / a
    u_m_b_a = u_m * b / a
    u_m_h_a = u_m * h / a
    u_m_r_a = u_m * r[:, np.newaxis] / a

    # Typo in paper for A_m, numerators sin terms are functions of u_m NOT v_m
    A_m = factor * (np.sin(u_m_c_a) - np.sin(u_m_b_a))
    A_m /= (u_m_h_a + np.sin(u_m_h_a)) * (iv(1, u_m) * u_m + iv(0, u_m)*tau_edge) / a # Eq 18
    # This result isn't in the paper direct but found by integrating Eq.9 over z = -h/2 -> h/2
    # Eq. 10 drops out due to sin asymmetry
    Z_ss_per_W = (material.dndT * (2*a/u_m * A_m * np.sin(u_m/a*h/2)) * iv(0, u_m_r_a)).sum(1)
    return Z_ss_per_W - abs(Z_ss_per_W).min()

def surface_deformation(r, a, b, c, h, material, barrel_material = None, T_ext=293.15, m_max = 10, root_accuracy=1e-12):
    """"Computes the surface thermo-elastic deformation of the HR surface per Watt
    absorbed ring heater power.


    Parameters
    ----------
    r : array_like
        Radial vector [m]
    a : float
        Mirror radius [m]
    b : float
        Ring heater lower boundary relative to center [m]
    c : float
        Ring heater upper boundary relative to center [m]
    h : float
        Mirror thickness [m]
    material : Material
        Mirror substrate material, see :py:mod:`finesse.materials`
    barrel_material : Material
        Barrel coating material, see :py:mod:`finesse.materials`
    T_ext : float, optional
        External temperature surrounding mirror
    m_max : int, optional
        Number of zeros to find, i.e. analytic order
    root_accuracy: float, optional
        Accuracy of root finding

    Returns
    -------
    U_z_rh_per_W : array_like
        1D radial vector of z displacement per Watt of ring heater power

    Notes
    -----
    See :cite:`T2500014` for derivation of these equations.
    """
    # checking for emissivity
    if barrel_material:
        emiss_face = material.emiss
        emiss_edge = barrel_material.emiss
    else:
        emiss_face = material.emiss
        emiss_edge = material.emiss

    tau_face = 4 * emiss_face * SIGMA_F * T_ext ** 3 * a / material.kappa
    tau_edge = 4 * emiss_edge * SIGMA_F * T_ext ** 3 * a / material.kappa
    u_m = zeros_u_m(a, h, tau_face, m_max, root_accuracy)
    v_m = zeros_v_m(a, h, tau_face, m_max, root_accuracy)

    factor = 1 / (material.kappa * np.pi * a * (c-b))
    u_m_c_a = u_m * c / a
    u_m_b_a = u_m * b / a
    u_m_h_a = u_m * h / a
    u_m_r_a = u_m * r[:, np.newaxis]  / a

    v_m_c_a = v_m * c / a
    v_m_b_a = v_m * b / a
    v_m_h_a = v_m * h / a
    v_m_r_a = v_m * r[:, np.newaxis] / a

    # Calculate A and B coeffiicients for temperature field, from T2500014, eq.4 & 5
    # https://dcc.ligo.org/LIGO-T2500014/public
    A_m = factor * (np.sin(u_m_c_a) - np.sin(u_m_b_a))
    A_m /= (u_m_h_a + np.sin(u_m_h_a))*(iv(1, u_m) * u_m / a + iv(0, u_m) * tau_edge / a)

    B_m = factor * (-np.cos(v_m_c_a) + np.cos(v_m_b_a))
    B_m /= (v_m_h_a - np.sin(v_m_h_a))*(iv(1, v_m) * v_m / a + iv(0, v_m) * tau_edge / a)

    # u_z:
    #  Compute only the z-independent part of T2500014,  eq.(10)
    u_z_factor = a * material.alpha * (material.poisson + 1 )
    u_z = u_z_factor * (- np.sin(u_m_h_a / 2) * iv(0, u_m_r_a) * A_m * v_m
                        - np.cos(v_m_h_a / 2) * iv(0, v_m_r_a) * B_m * u_m
                        + np.sin(u_m_h_a / 2) * A_m * v_m
                        + np.cos(v_m_h_a / 2) * B_m * u_m)
    u_z /= (u_m * v_m)
    u_z = u_z.sum(1)

    # du_z:
    # Compute only dz term that's only dependent on r( eq. 16)
    du_z_factor =  6 * a * material.alpha  * (material.poisson - 1) / h ** 3
    du_z = du_z_factor  *  r[:, np.newaxis] ** 2 * ( 2 * a * np.sin(v_m_h_a / 2) - h * v_m * np.cos(v_m_h_a / 2)) * iv(1, v_m) * B_m
    du_z /= v_m ** 3
    du_z = du_z.sum(1)

    U_z_rh_per_W = u_z + du_z
    U_z_rh_per_W -= U_z_rh_per_W.max()
    U_z_rh_per_W *= -1  # HR surface has positive z displacement

    return U_z_rh_per_W


def substrate_deformation_depth(r, z, a, b, c, h, material, barrel_material = None, T_ext=293.15, m_max = 10, root_accuracy=1e-12):
    """
    Computes the depth displacements throughout the bulk of an optic due to 1W
    absorption of ring heater. Displacement is in units of m per absorbed Watts.

    Parameters
    ----------
    r : array_like
        Radial vector [m]
    z  :array_like
        Longitudinal points, should sample points between -h/2 and h/2.
    a : float
        Mirror radius [m]
    b : float
        Ring heater lower boundary relative to center [m]
    c : float
        Ring heater upper boundary relative to center [m]
    h : float
        Mirror thickness [m]
    material : :class:`finesse.materials.Material`
        Mirror substrate material, see :py:mod:`finesse.materials`
    barrel_material : :class:`finesse.materials.Material`
        Barrel coating material, see :py:mod:`finesse.materials`
    T_ext : float, optional
        External temperature surrounding mirror
    m_max : int, optional
        Number of zeros to find, i.e. analytic order
    root_accuracy: float, optional
        Accuracy of root finding

    Returns
    -------
    U_z_rh_per_W : ndarray(shape=(z.size, r.size))
        Array of z displacements throughout the substrate
        per absorbed Watts of ring heater

    Notes
    -----
    See :cite:`T2500014` for derivation of these equations.
    """
    if barrel_material:
        emiss_face = material.emiss
        emiss_edge = barrel_material.emiss
    else:
        emiss_face = material.emiss
        emiss_edge = material.emiss

    z = z[:, np.newaxis, np.newaxis]
    r = r[:, np.newaxis]

    tau_face = 4 * emiss_face * SIGMA_F * T_ext ** 3 * a / material.kappa
    tau_edge = 4 * emiss_edge * SIGMA_F * T_ext ** 3 * a / material.kappa
    u_m = zeros_u_m(a, h, tau_face, m_max, root_accuracy)
    v_m = zeros_v_m(a, h, tau_face, m_max, root_accuracy)

    factor = 1 / (material.kappa * np.pi * a * (c-b))
    u_m_c_a = u_m * c / a
    u_m_b_a = u_m * b / a
    u_m_h_a = u_m * h/  a
    u_m_r_a = u_m * r / a
    u_m_z_a = u_m * z / a

    v_m_c_a = v_m * c / a
    v_m_b_a = v_m * b / a
    v_m_h_a = v_m * h / a
    v_m_r_a = v_m * r/  a
    v_m_z_a = v_m * z / a

    # Calculate A and B coeffiicients for temperature field, from T2500014, eq.4 & 5
    # https://dcc.ligo.org/LIGO-T2500014/public
    A_m = factor * (np.sin(u_m_c_a) - np.sin(u_m_b_a))
    A_m /= (u_m_h_a + np.sin(u_m_h_a))*(iv(1, u_m) * u_m / a + iv(0, u_m) * tau_edge / a)

    B_m = factor * (-np.cos(v_m_c_a) + np.cos(v_m_b_a))
    B_m /= (v_m_h_a - np.sin(v_m_h_a))*(iv(1, v_m) * v_m / a + iv(0, v_m) * tau_edge / a)

    #  Compute u_z, T2500014  eq.(10)
    u_z_factor = a * material.alpha * (material.poisson + 1 )
    u_z = u_z_factor * (  np.sin(u_m_h_a / 2) * A_m * v_m
                        + np.sin(u_m_z_a) * iv(0, u_m_r_a) * A_m * v_m
                        + np.cos(v_m_h_a / 2) * B_m * u_m
                        - np.cos(v_m_z_a) * iv(0, v_m_r_a) * B_m * u_m
                        )
    u_z /= (u_m * v_m)
    u_z = u_z.sum(2)

    # COmpute du_z,T2500014, eq. 7, 15, 16
    du_z_1_factor = 2 * a * material.alpha *  material.poisson / h ** 3
    du_z_1 = du_z_1_factor * ( 2 * h ** 2 * np.sin(u_m_h_a / 2) * iv(1, u_m) * A_m * v_m ** 3
                              + 6 * z * (2  * a * np.sin(v_m_h_a / 2) -
                                                        h * np.cos(u_m_h_a / 2) * v_m) *
                                                        B_m * u_m ** 2
                              )
    du_z_1 /= (u_m ** 3 *  v_m ** 3)

    du_z_2_factor  =  6 * a * material.alpha  * (material.poisson - 1) / h ** 3
    du_z_2 =  du_z_2_factor  *  r** 2 * ( 2 * a * np.sin(v_m_h_a / 2)
                                                       - h * v_m * np.cos(v_m_h_a / 2)) * iv(1, v_m) * B_m

    du_z_2 /= v_m ** 3
    du_z = du_z_1 + du_z_2
    du_z = du_z.sum(2)

    U_z_rh_per_W = u_z + du_z
    U_z_rh_per_W *= -1 # HR surface has positive z displacement

    return U_z_rh_per_W



import numpy as np
import pytest

import finesse
import finesse.components

from finesse.detectors import FieldDetector
from finesse.exceptions import BeamTraceException


@pytest.fixture
def telescope_model():
    model = finesse.Model()
    laser1 = model.add(finesse.components.Laser("laser1"))
    telescope = model.add(finesse.components.Telescope("telescope"))
    laser2 = model.add(finesse.components.Laser("laser2"))
    model.link(laser1, telescope, laser2)
    model.modes("even", maxtem=2)
    return model


def test_telescope_fields(telescope_model):
    # ensure fields transmit with a mode-mismatch unperturbed
    telescope_model.laser1.p1.o.q = finesse.BeamParam(w0=0.1, z=0.1)
    telescope_model.laser2.p1.o.q = finesse.BeamParam(w0=0.2, z=0.2)
    telescope_model.laser2.set_output_field([0, 1, 0], telescope_model.homs)

    telescope_model.add(FieldDetector("E1", telescope_model.telescope.p1.o, 0))
    telescope_model.add(FieldDetector("E2", telescope_model.telescope.p2.o, 0))
    out = telescope_model.run()

    assert np.allclose(out["E2"], [1, 0, 0])
    assert np.allclose(abs(out["E1"]), [0, 1, 0])


def test_telescope_detect_mismatches(telescope_model):
    telescope_model.laser1.p1.o.q = finesse.BeamParam(w0=0.1, z=0.1)
    telescope_model.laser2.p1.o.q = finesse.BeamParam(w0=0.2, z=0.2)
    assert len(telescope_model.detect_mismatches()) == 0


def test_telescope_trace_error(telescope_model):
    telescope_model.laser1.p1.o.q = finesse.BeamParam(w0=0.1, z=0.1)

    with pytest.raises(BeamTraceException):
        # missing beam parameters on laser2 side
        telescope_model.beam_trace()



"""Functions for computing outputs for higher order aberrations of spherical optical
components."""

import numpy as np


def spherical_surface(X, Y, R, degrees=0):
    """Computes the spherical height (sagitta) for a spherical mirror normal incidence.
    Will return NaNs for regions in X**2+Y**2 > R**2.

    Parameters
    ----------
    X, Y : [array_like| float]
        Coordinates to compute surface at
    R : float
        Radius of curvature
    degrees : float
        Angle of incidence in degrees

    Returns
    -------
    Z : [array_like|float]
        Height of the spherical surface

    Notes
    -----
    Coorindate system used is left-handed with positive R being in the +z direction.
    Formula is from equation 29 in :cite:`Hello_1996`.
    """
    sign = np.sign(R)
    R = abs(R)
    X_sq = X ** 2
    Y_sq = Y ** 2

    if degrees != 0:
        radians = np.radians(degrees)
        cos = np.cos(radians)
        sin = np.sin(radians)
        cos2 = np.cos(2 * radians)
        sin2 = np.sin(2 * radians)
        f_xy = R * cos - np.sqrt(R ** 2 * cos ** 2 - 2 * R * X * sin - X_sq - Y_sq)
        g_xy = np.sqrt(cos ** 2 - 2 * X * sin / R - (X_sq + Y_sq) / R ** 2)
        numer = X * sin2 - f_xy * cos2
        denom = 2 * sin2 * g_xy * (sin + X / R) - cos2 * (1 - 2 * g_xy ** 2)
        s_xy = sign * (f_xy - numer / denom)
    else:
        # simplify angle = 0
        f_xy = R - np.sqrt(R ** 2 - X_sq - Y_sq)
        g_xy = np.sqrt(1 - (X_sq + Y_sq) / R ** 2)
        s_xy = sign * f_xy * (1 - 1 / (1 - 2 * g_xy ** 2))
    return s_xy / 2  # includes factor of 2 on reflection



"""Test parsing of beam property detectors."""

import pytest
from finesse.script import parse
from finesse.detectors.bpdetector import BP_KEYWORDS


@pytest.mark.parametrize("keyword,expected", BP_KEYWORDS.items())
def test_beam_property_detector(keyword, expected):
    """Test that beam property detector properties parse correctly."""
    model = parse(
        f"""
        m m1 R=0.99 T=0.01
        bp m1_w m1.p2.o {keyword}
        """
    )

    assert model.m1_w.prop == expected



"""Model component network unit test fixtures."""

import pytest
from finesse import Model
from finesse.components import Laser, Beamsplitter, Mirror, Cavity
from finesse.detectors import PowerDetector


@pytest.fixture(scope="module")
def network_model_michelson():
    """Fixture with a Michelson model to test component network generation."""
    model = Model()

    model.add(
        [
            Laser("L0"),
            Beamsplitter("BS"),
            Mirror("IMX", Rc=-0.6),
            Mirror("IMY", Rc=-0.6),
            Mirror("EMX", Rc=0.6),
            Mirror("EMY", Rc=0.6),
        ]
    )
    model.connect(model.L0.p1, model.BS.p1, L=1)
    model.connect(model.BS.p2, model.IMY.p1, L=1)
    model.connect(model.IMY.p2, model.EMY.p1, L=1)
    model.connect(model.BS.p3, model.IMX.p1, L=1)
    model.connect(model.IMX.p2, model.EMX.p1, L=1)

    model.add(PowerDetector("refl", model.IMX.p1.o))
    model.add(Cavity("cav_x", model.IMX.p2.o, model.IMX.p2.i))
    model.add(Cavity("cav_y", model.IMY.p2.o, model.IMY.p2.i))

    return model, model.to_component_network()


@pytest.fixture(scope="module")
def network_model_sagnac():
    """Fixture with a Sagnac model to test component network generation."""
    model = Model()

    model.add(
        [
            Laser("L0"),
            Beamsplitter("BS"),
            Mirror("M1", Rc=-1.7),
            Mirror("M2"),
            Mirror("M3", Rc=1.7),
        ]
    )
    model.connect(model.L0.p1, model.BS.p1, L=1)
    model.connect(model.BS.p2, model.M1.p1, L=1)
    model.connect(model.M1.p2, model.M2.p1, L=1)
    model.connect(model.M2.p2, model.M3.p1, L=1)
    model.connect(model.M3.p2, model.BS.p3, L=1)

    model.add(PowerDetector("trns", model.BS.p4.o))

    return model, model.to_component_network()



import numpy as np
from copy import deepcopy
from finesse import constants


cpdef rebuild_model_settings(d) :
    """
    Rebuilds the model settings from a dictionary.

    Parameters
    ----------
    d : dict
        A dictionary where each key-value pair represents a setting name and its value.

    Returns
    -------
    ModelSettings
        A new `ModelSettings` object with its attributes set according to `d`.

    """
    s = ModelSettings()
    for k,v in d.items():
        setattr(s, k, v)
    return s


cdef class ModelSettings:
    """
    A class used to represent the settings of a model.

    Attributes
    ----------
    phase_config : PhaseConfig
        The configuration of the phase.
    homs_view : ndarray
        A view on the higher order modes (HOMs).
    max_n : int
        The maximum n value in the HOMs.
    max_m : int
        The maximum m value in the HOMs.
    num_HOMs : int
        The number of HOMs.
    lambda0 : float
        The wavelength.
    f0 : float
        The frequency, calculated as the speed of light divided by the wavelength.
    k0 : float
        The wave number, calculated as 2*pi divided by the wavelength.

    Methods
    -------
    __reduce__()
        Returns a tuple that can be used to rebuild the object using `rebuild_model_settings`.
    set_lambda0(value)
        Sets the wavelength and updates the frequency and wave number accordingly.
    """

    def __init__(self):
        self.phase_config = PhaseConfig()
        self.homs_view = None
        self.max_n = 0
        self.max_m = 0
        self.num_HOMs = 0
        self.set_lambda0(1064e-9)

    def __reduce__(self):
        """
        Returns a tuple that can be used to rebuild the object using `rebuild_model_settings`.

        Returns
        -------
        tuple
            A tuple containing the `rebuild_model_settings` function and a dictionary of the object's attributes.
        """
        # This is all to get around that memoryviews can't be deepcopied
        d = {"homs_view": np.asarray(self.homs_view)}
        d.update(
            {
                _: deepcopy(getattr(self, _))
                for _ in dir(self) if not _.startswith("__") and _ not in d and not callable(getattr(self, _))
            }
        )
        return rebuild_model_settings, (d,)

    @property
    def homs(self):
        """
        Returns a view on the higher order modes (HOMs).

        Returns
        -------
        ndarray
            A view on the HOMs.
        """
        return np.asarray(self.homs_view)

    @homs.setter
    def homs(self, value):
        """
        Sets the HOMs and updates the maximum n and m values and the number of HOMs.

        Parameters
        ----------
        value : ndarray
            The new HOMs.
        """
        self.homs_view = value
        self.max_n    = np.max(self.homs_view[:,0])
        self.max_m    = np.max(self.homs_view[:,1])
        self.num_HOMs = self.homs_view.shape[0]

    def set_lambda0(self, value):
        """
        Sets the wavelength and updates the frequency and wave number accordingly.

        Parameters
        ----------
        value : float
            The new wavelength.
        """
        self.lambda0 = value
        self.f0 = constants.C_LIGHT / self.lambda0
        self.k0 = 2.0 * constants.PI / self.lambda0


cdef class PhaseConfig:
    # TODO maybe this can be removed, not sure if it's used anywhere
    # other than in the model settings
    pass



"""This test ensures the output of `run_fsig_sweep` performed by the `FrequencyResponse`
action is independent of `x_scale`."""
from finesse import Model
from finesse.analysis.actions import FrequencyResponse

import numpy as np

import pytest


# prepare some models
@pytest.fixture()
def swinging_mirror_no_filter():
    model = Model()
    model.parse(
        """
        m m1 R=1 T=0

        pendulum m1_sus m1.mech mass=1

        fsig(1)
        """
    )
    return model


@pytest.fixture()
def swinging_mirror_with_e_filter():
    model = Model()
    model.parse(
        """
        fsig(1)

        m m1 R=1 T=0
        zpk ctrl [] [] 1 gain=1
        pendulum m1_sus m1.mech mass=1

        link(ctrl, m1.mech.F_z)
        """
    )
    return model


@pytest.fixture()
def swinging_mirror_with_m_filter():
    model = Model()
    model.parse(
        """
        fsig(1)

        m m1 R=1 T=0
        pendulum m1_sus m1.mech mass=1
        zpk filter [] [] 1 gain=1

        link(m1.mech.z, filter)
        """
    )
    return model


@pytest.fixture()
def swinging_mirror_with_both_filters():
    model = Model()
    model.parse(
        """
        m m1 R=1 T=0

        zpk ctrl [] [] 1 gain=1
        pendulum m1_sus m1.mech mass=1
        link(ctrl, m1.mech.F_z)

        zpk filter [] [] 1 gain=1
        link(m1.mech.z, filter)

        fsig(1)
        """
    )
    return model


# prepare comparisons
def get_outs(model, inject_into, read_out_of, open_loop=False):
    """Runs and returns two FRs with different x_scale values."""

    f = np.logspace(-1, 6, 11)

    model1 = model.deepcopy()
    model1._settings.x_scale = 1e-9
    out1 = model1.run(
        FrequencyResponse(f, inject_into, read_out_of, open_loop=open_loop)
    )

    model2 = model.deepcopy()
    model2._settings.x_scale = 1
    out2 = model2.run(
        FrequencyResponse(f, inject_into, read_out_of, open_loop=open_loop)
    )

    return out1.out, out2.out


# run some tests
#
# These tests run through every possible input/output signal type combination
#   for both closed loop and open loop to ensure scaling is working.
#
# 1) m to m
# 2) m to m (with a zpk present)
# 3) e to m
# 4) e to e
# 5) m to e


# m to m
def test_cl_frequency_response_x_scale_m_to_m(swinging_mirror_no_filter):
    """Test effect of changing x_scale on mech to mech frequency response in closed
    loop."""

    # closed loop
    out1, out2 = get_outs(swinging_mirror_no_filter, "m1.mech.F_z", "m1.mech.z")
    assert abs(out1 - out2).all() <= 1e-15


def test_ol_frequency_response_x_scale_m_to_m(swinging_mirror_no_filter):
    """Test effect of changing x_scale on mech to mech frequency response in open
    loop."""

    # open loop
    out1, out2 = get_outs(
        swinging_mirror_no_filter, "m1.mech.F_z", "m1.mech.z", open_loop=True
    )
    assert abs(out1 - out2).all() <= 1e-15


# m to m w/ e filter
def test_cl_frequency_response_x_scale_m_to_m_w_e_filter(swinging_mirror_with_e_filter):
    """Test effect of changing x_scale on mech to mech frequency response in closed
    loop."""

    # closed loop
    out1, out2 = get_outs(swinging_mirror_with_e_filter, "m1.mech.F_z", "m1.mech.z")
    assert abs(out1 - out2).all() <= 1e-15


def test_ol_frequency_response_x_scale_m_to_m_w_e_filter(swinging_mirror_with_e_filter):
    """Test effect of changing x_scale on mech to mech frequency response in open
    loop."""

    # open loop
    out1, out2 = get_outs(
        swinging_mirror_with_e_filter, "m1.mech.F_z", "m1.mech.z", open_loop=True
    )
    assert abs(out1 - out2).all() <= 1e-15


# e to m
def test_cl_frequency_response_x_scale_e_to_m(swinging_mirror_with_e_filter):
    """Test effect of changing x_scale on elec to mech frequency response in closed
    loop."""

    # closed loop
    out1, out2 = get_outs(swinging_mirror_with_e_filter, "ctrl.p1", "m1.mech.z")
    assert abs(out1 - out2).all() <= 1e-15


def test_ol_frequency_response_x_scale_e_to_m(swinging_mirror_with_e_filter):
    """Test effect of changing x_scale on elec to mech frequency response in open
    loop."""

    # open loop
    out1, out2 = get_outs(
        swinging_mirror_with_e_filter, "ctrl.p1", "m1.mech.z", open_loop=True
    )
    assert abs(out1 - out2).all() <= 1e-15


# e to e
def test_cl_frequency_response_x_scale_e_to_e(swinging_mirror_with_both_filters):
    """Test effect of changing x_scale on elec to mech frequency response in closed
    loop."""

    # closed loop
    out1, out2 = get_outs(swinging_mirror_with_both_filters, "ctrl.p1", "filter.p2")
    assert abs(out1 - out2).all() <= 1e-15


def test_ol_frequency_response_x_scale_e_to_e(swinging_mirror_with_both_filters):
    """Test effect of changing x_scale on elec to mech frequency response in open
    loop."""

    # open loop
    out1, out2 = get_outs(
        swinging_mirror_with_both_filters, "ctrl.p1", "filter.p2", open_loop=True
    )
    assert abs(out1 - out2).all() <= 1e-15


# m to e
def test_cl_frequency_response_x_scale_m_to_e(swinging_mirror_with_m_filter):
    """Test effect of changing x_scale on mech to mech frequency response in closed
    loop."""

    # closed loop
    out1, out2 = get_outs(swinging_mirror_with_m_filter, "m1.mech.F_z", "filter.p2")
    assert abs(out1 - out2).all() <= 1e-15


def test_ol_frequency_response_x_scale_m_to_e(swinging_mirror_with_m_filter):
    """Test effect of changing x_scale on mech to mech frequency response in open
    loop."""

    # open loop
    out1, out2 = get_outs(
        swinging_mirror_with_m_filter, "m1.mech.F_z", "filter.p2", open_loop=True
    )
    assert abs(out1 - out2).all() <= 1e-15



import pytest
import numpy as np


@pytest.mark.parametrize("initial_z0", (+1200, -1200))
def test_initial_beam_param(initial_z0):
    """Test that an initial beam parameter is overwritten by xaxis.

    This test is based on the discussion in
    https://chat.ligo.org/ligo/pl/rzm4ews487dzuyqmsrzwj5jfqa
    where a beam parameter initially set
    was not being overwitten by the xaxis.

    This test failed in 86bea1a4
    and was fixed in 10af113c.
    """
    import finesse

    out = []
    for z0 in 0, initial_z0:
        script = f"""
        laser l1 P=1.5

        gauss g1 l1.p1.o w0=10e-3 z={z0}
        modes(maxtem=7)

        space s1 l1.p1 bs1.p1 L=1000
        beamsplitter bs1 R=1 T=0 alpha=30 xbeta=1e-5
        space s2 bs1.p2 bs2.p1 L=400
        beamsplitter bs2 R=1 T=0 alpha=30 xbeta=bs1.xbeta
        space s3 bs2.p2 n1.p1 L=600
        nothing n1

        power_detector_dc pd1 node=n1.p1.i pdtype=none

        xaxis(g1.zx, lin, 0, -2000, 60)
        """
        kat = finesse.Model()
        kat.parse(script)
        kat.g1.zy = kat.g1.zx.ref
        out.append(kat.run())

    assert np.all(out[0]["pd1"] == out[1]["pd1"])



from finesse.cymath.complex cimport complex_t, cnorm, cabs, cimag, creal, conj

cdef extern from "math.h" nogil:
    double atan2(double y, double x)
    double fabs(double arg)
    double sqrt(double arg)

cdef extern from "constants.h":
    long double PI


### Low-level beam parameter struct and functions ###

cdef struct beam_param:
    complex_t q
    # TODO (sjr) Should probably change below to pointers to
    # refractive index and wavelength because these get initialised
    # in a simulation via the connector workspaces and model data
    # respectively so could just keep addresses here
    double nr
    double wavelength

cdef inline double bp_beamsize(const beam_param* bp) noexcept nogil:
    return beam_size(bp.q, bp.nr, bp.wavelength)

cdef inline double bp_waistsize(const beam_param* bp) noexcept nogil:
    return waist_size(bp.q, bp.nr, bp.wavelength)

cdef inline double bp_waistpos(const beam_param* bp) noexcept nogil:
    return creal(bp.q)

cdef inline double bp_rayleigh(const beam_param* bp) noexcept nogil:
    return cimag(bp.q)

cdef inline double bp_gouy(const beam_param* bp) noexcept nogil:
    return gouy(bp.q)

cdef inline double bp_divergence(const beam_param* bp) noexcept nogil:
    return divergence(bp.q, bp.nr, bp.wavelength)

cdef inline double bp_radius_curvature(const beam_param* bp) noexcept nogil:
    return radius_curvature(bp.q)

cdef inline double bp_defocus(const beam_param* bp) noexcept nogil:
    return defocus(bp.q)


cdef inline double bp_overlap(const beam_param* bp_x, const beam_param* bp_y) noexcept nogil:
    return overlap(bp_x.q, bp_y.q)


### Corresponding functions acting on standard complex_t data ###

cpdef inline double beam_size(complex_t q, double nr, double lambda0) noexcept nogil:
    return cabs(q) * sqrt(lambda0 / (nr * PI * cimag(q)))

cpdef inline double waist_size(complex_t q, double nr, double lambda0) noexcept nogil:
    return sqrt(lambda0 * cimag(q) / (PI * nr))

cpdef inline double gouy(complex_t q) noexcept nogil:
    return atan2(creal(q), cimag(q))

cpdef inline double divergence(complex_t q, double nr, double lambda0) noexcept nogil:
    return lambda0 / (waist_size(q, nr, lambda0) * PI)

cpdef inline double radius_curvature(complex_t q) noexcept nogil:
    cdef:
        double z = creal(q)
        double zr = cimag(q)
        double zr_on_z = zr / z

    return z * (1 + zr_on_z * zr_on_z)

cpdef inline double defocus(complex_t q) noexcept nogil:
    cdef:
        double z = creal(q)
        double zr = cimag(q)

    return z / (z * z + zr * zr)


cpdef complex_t transform_q(
    const double[:, ::1] M,
    complex_t q1,
    double nr1,
    double nr2
) noexcept nogil

cdef complex_t c_transform_q(
    const double* M,
    complex_t q1,
    double nr,
    double nr2
) noexcept nogil

cpdef complex_t inv_transform_q(
    const double[:, ::1] M,
    complex_t q2,
    double nr1,
    double nr2
) noexcept nogil


cpdef inline double overlap(complex_t qx, complex_t qy) noexcept nogil:
    return fabs(4 * cimag(qx) * cimag(qy)) / cnorm(conj(qx) - qy)


cdef void c_abcd_multiply(
    const double* m1,
    const double* m2,
    double* out,
) noexcept nogil

cdef void abcd_multiply(
    const double[:, ::1] m1,
    const double[:, ::1] m2,
    double[:, ::1] out,
) noexcept nogil

cpdef void sym_abcd_multiply(
    object[:, ::1] m1,
    object[:, ::1] m2,
    object[:, ::1] out,
) noexcept

cpdef bint is_abcd_changing(object[:, ::1] M) noexcept



import numpy as np
from finesse.utilities.polyfit import (
    polyfit2d,
    polyfit2d_eval,
    polyfit2d_indices,
    polyfit2d_index,
)


def test_fit_unity_coefficients():
    x = np.linspace(-1, 1, num=201)
    y = np.linspace(-1, 1, num=201)
    kx = ky = 5
    X, Y = np.meshgrid(x, y)
    true_coeffs = np.ones((kx + 1, ky + 1))
    Z = polyfit2d_eval(x, y, true_coeffs, kx, ky)
    fit_coeffs, _, _, _ = polyfit2d(x, y, Z, kx, ky)
    assert np.all(abs(fit_coeffs - true_coeffs.ravel()) < 1e-12)


def test_indices():
    kx = ky = 2
    indices = polyfit2d_indices(kx, ky)
    assert np.all(
        indices
        == np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )
    )


def test_index():
    kx = 2
    ky = 2
    indices = polyfit2d_indices(kx, ky)
    for index, (a, b) in enumerate(indices):
        assert polyfit2d_index(kx, a, b) == index



"""Optical components performing modulation of beams."""

import logging
import finesse
import numpy as np

from finesse.tracing import abcd
from finesse.components.general import InteractionType, Connector, FrequencyGenerator
from finesse.components.node import NodeType, NodeDirection
from finesse.parameter import (
    float_parameter,
    int_parameter,
    bool_parameter,
    enum_parameter,
    ParameterState,
)
from finesse.symbols import Variable, Matrix
from finesse.enums import ModulatorType

LOGGER = logging.getLogger(__name__)


@float_parameter("f", "Frequency", units="Hz")
@float_parameter("midx", "Modulation index")
@float_parameter("phase", "Phase", units="degrees")
@int_parameter(
    "order",
    "Maximum modulation order",
    changeable_during_simulation=False,
    validate="_check_order",
)
@enum_parameter(
    "mod_type",
    f"Modulation type {tuple(_ for _ in ModulatorType.__members__.keys())}",
    ModulatorType,
    changeable_during_simulation=False,
    validate="_check_mod_type",
)
@bool_parameter("positive_only", "Positive only", changeable_during_simulation=False)
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Modulator(Connector, FrequencyGenerator):
    """Represents a modulator optical component with associated properties such as
    modulation frequency, index and order.

    Parameters
    ----------
    name : str
        Name of newly created modulator.

    f : float or :class:`.Frequency`, optional
        Frequency of the modulation (in Hz) or :class:`.Frequency` object.

    midx : float
        Modulation index, >= 0.

    order : int, optional
        Maximum order of modulations to produce. Must be 1 for
        amplitude modulation. Defaults to 1.

    mod_type : str, optional
        Modulation type, either 'am' (:ref:`amplitude
        modulation<amp_mod>`) or 'pm' (:ref:`phase
        modulation<phase_mod>`). Defaults to 'pm'.

    phase : float, optional
        Relative phase of modulation (in degrees). Defaults to 0.0.

    positive_only : bool, optional
        If True, only produce positive-frequency sidebands. Defaults to False.
    """

    def __init__(
        self,
        name,
        f,
        midx,
        order=1,
        mod_type=ModulatorType.pm,
        phase=0.0,
        positive_only=False,
    ):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)

        self.f = f
        self.midx = midx
        self.phase = phase
        self.order = order
        self.mod_type = mod_type
        self.positive_only = positive_only

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

        self._register_node_coupling("amp_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("amp_P2o", self.amp.i, self.p2.o)

    def optical_equations(self):
        settings = self._model._settings

        with finesse.symbols.simplification():
            if self.mod_type == ModulatorType.pm:
                jv = finesse.symbols.FUNCTIONS["jv"]
                _k_ = Variable("_k_")  # modulation order of the operator

                # Scalar for a phase modulation operator
                scalar = (1j) ** _k_ * jv(_k_, self.midx.ref)

                if settings.is_modal:
                    return {
                        f"{self.name}.P1i_P2o": scalar * Matrix(f"{self.name}.K12"),
                        f"{self.name}.P2i_P1o": scalar * Matrix(f"{self.name}.K21"),
                    }
                else:
                    return {
                        f"{self.name}.P1i_P2o": scalar,
                        f"{self.name}.P2i_P1o": scalar,
                    }
            else:
                raise NotImplementedError()

    def _modulation_frequencies(self):
        order = self.order.eval()
        if order - int(order) != 0:
            raise ValueError(
                f"Modulation order for {self.name} must be an integer not {order}"
            )
        order = int(order)
        orders = list(range(-order, 1 + order))
        orders.pop(order)  # remove 0
        fm = np.dot(self.f.ref, orders)
        return fm

    def _couples_frequency(self, ws, connection, f_in, f_out):
        if connection in ("P2i_P1o", "P1i_P2o"):
            if f_in == f_out:
                return True
            elif (f_in, f_out) in ws.carrier_frequency_couplings:
                return True
            elif (f_out, f_in) in ws.carrier_frequency_couplings:
                return True
            elif (f_in, f_out) in ws.signal_frequency_couplings:
                return True
            elif (f_out, f_in) in ws.signal_frequency_couplings:
                return True
            else:
                return False
        else:  # any other signal connection
            return True

    def _check_order(self, value):
        if self.mod_type == ModulatorType.am and int(value) != 1:
            raise ValueError(
                "Modulation order must be 1 when using amplitude modulation"
            )
        return value

    def _check_mod_type(self, value):
        try:
            value = ModulatorType(value)
        except ValueError:
            try:
                value = ModulatorType[value]  # cast input into enum
            except KeyError:
                raise ValueError(
                    f"'{value}' is not a valid Modulator type, options are {tuple(_ for _ in ModulatorType.__members__.keys())}"
                )

        if self.order.state == ParameterState.Numeric:
            if value == ModulatorType.am and self.order > 1:
                raise ValueError("Amplitude modulation can only be used with order 1")
        return value

    @property
    def f(self):
        """Source frequency representing modulation.

        Returns
        -------
        :class:`float` or :class:`.Frequency`
            The modulation frequency.
        """
        return self.__f

    @f.setter
    def f(self, value):
        self.__f = value

    @property
    def order(self):
        """The maximum order of modulations produced by the modulator.

        Returns
        -------
        int
            The maximum modulation order.
        """
        return self.__order

    @order.setter
    def order(self, value):
        self.__order = int(value)

    def _resymbolise_ABCDs(self):
        M_sym = abcd.none()
        for direction in ["x", "y"]:
            # Matrices same for both node couplings
            self.register_abcd_matrix(
                M_sym,
                (self.p1.i, self.p2.o, direction),
                (self.p2.i, self.p1.o, direction),
            )

    def _get_workspace(self, sim):
        from finesse.components.modal.modulator import (
            ModulatorWorkspace,
            modulator_carrier_fill,
            modulator_signal_optical_fill,
            modulator_signal_phase_fill,
            modulator_signal_amp_fill,
        )

        ws = ModulatorWorkspace(self, sim)
        changing = (
            self.midx.is_changing
            or self.phase.is_changing
            or self.f.is_changing
            or sim.is_component_in_mismatch_couplings(self)
            or self in sim.trace_forest
            or sim.carrier.any_frequencies_changing
        )

        ws.carrier.add_fill_function(
            modulator_carrier_fill,
            changing,
        )

        if sim.signal:
            # modulator for signals doesn't actually need refilling each time if
            # signal frequency is changing, as the coupling depends on the
            ws.signal.add_fill_function(
                modulator_signal_optical_fill,
                self.midx.is_changing or self.phase.is_changing,
            )

            if ws.amp_signal_enabled:
                # Amplitude signal elements do not depend on the signal frequency
                ws.signal.add_fill_function(
                    modulator_signal_amp_fill,
                    self.midx.is_changing or self.phase.is_changing,
                )

            if ws.phs_signal_enabled:
                # Phase signal elements do not depend on the signal frequency
                ws.signal.add_fill_function(
                    modulator_signal_phase_fill,
                    self.midx.is_changing or self.phase.is_changing,
                )
        return ws



import finesse
import numpy as np


def test_model_parameter_extract_noxaxis():
    model = finesse.script.parse(
        """
    l l1 P=2.3
    mathd P l1.P
    mathd P_10 10*l1.P
    """
    )
    out = model.run()

    assert out["P"] == model.l1.P.value
    assert out["P_10"] == model.l1.P.value * 10


def test_model_parameter_extract_xaxis():
    model = finesse.script.parse(
        """
    l l1 P=2.3
    mathd P l1.P
    mathd P_10 10*l1.P
    xaxis(l1.P, lin, 0, 1, 3)
    """
    )
    out = model.run()

    assert np.allclose(out["P"].astype(float), out.x1)
    assert np.allclose(out["P_10"].astype(float), out.x1 * 10)


def test_power_detector():
    model = finesse.script.parse(
        """
    l l1
    pd Y l1.p1.o
    mathd Z1 Y
    mathd Z2 Y*10
    xaxis(l1.P, lin, 0, 1, 3)
    """
    )
    out = model.run()

    assert np.allclose(out["Z1"].astype(float), out.x1)
    assert np.allclose(out["Z2"].astype(float), out.x1 * 10)


def test_power_detector_1_fsig():
    model = finesse.script.parse(
        """
    l l1
    pd1 Y l1.p1.o f=fsig
    sgen sg l1.pwr.i phase=90
    mathd Z1 Y
    mathd Z2 Y*10
    xaxis(fsig, log, 1, 10, 3)
    fsig(1)
    """
    )
    out = model.run()
    assert np.allclose(out["Z1"].astype(complex), 1j * np.ones_like(out.x1))
    assert np.allclose(out["Z2"].astype(complex), 1j * np.ones_like(out.x1) * 10)


def test_noise_detector():
    model = finesse.script.parse(
        """
    l l1
    qnoised Y l1.p1.o
    mathd Z1 Y
    mathd Z2 Y*10
    xaxis(fsig, log, 1, 10, 3)
    fsig(1)
    """
    )
    out = model.run()
    hf = model.f0 * finesse.constants.H_PLANCK

    assert np.allclose(out["Z1"].astype(float) ** 2 / hf, 2 * np.ones_like(out.x1))
    assert np.allclose(
        out["Z2"].astype(float) ** 2 / hf, 2 * np.ones_like(out.x1) * 100
    )



"""Laser-type optical components for producing beams."""

import logging
import math
import types

import numpy as np
from finesse.env import warn

from .general import Connector, FrequencyGenerator, NoiseType, LocalDegreeOfFreedom
from .node import NodeType, NodeDirection

from ..cymath.complex import crotate
from ..parameter import float_parameter, bool_parameter

LOGGER = logging.getLogger(__name__)


@float_parameter("P", "Power", units="W")
@float_parameter("phase", "Phase", units="degrees")
@float_parameter("f", "Frequency", units="Hz")
@bool_parameter("signals_only", "Signals only", changeable_during_simulation=False)
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Laser(Connector, FrequencyGenerator):
    """Represents a laser producing a beam with associated properties such as power and
    frequency.

    Parameters
    ----------
    name : str
        Name of the newly created laser.

    P : float, optional
        Power of the laser (in Watts), defaults to 1 W.

    f : float or :class:`.Frequency`, optional
        Frequency-offset of the laser from the default (in Hz) or :class:`.Frequency`
        object. Defaults to 0 Hz offset.

    phase : float, optional
        Phase-offset of the laser from the default, defaults to zero.

    signals_only : bool, optional
        When True, this laser component will only inject signal sidebands. They will use
        the current carrier value as a scaling terms but the carrier will not be
        injected into the simulation. This allows a user to just inject signal sidebands
        into a model.

    Attributes
    ----------
    add_gouy_phase : bool
        When set to True the gouy phase of the current beam parameters values at the
        laser will be added to the optical field outputs during the simulation. When
        False, it will not. This can be used with :meth:`.set_output_field` to force a
        particular optical field output from a laser.
    """

    DEFAULT_POWER_COEFFS = {(0, 0): (1.0, 0.0)}

    def __init__(self, name, P=1, f=0, phase=0, signals_only=False):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        # Modulation inputs
        self._add_port("pwr", NodeType.ELECTRICAL)
        self.pwr._add_node("i", NodeDirection.INPUT)
        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)
        self._add_port("frq", NodeType.ELECTRICAL)
        self.frq._add_node("i", NodeDirection.INPUT)
        self._add_port("dx", NodeType.ELECTRICAL)
        self.dx._add_node("i", NodeDirection.INPUT)
        self._add_port("dy", NodeType.ELECTRICAL)
        self.dy._add_node("i", NodeDirection.INPUT)
        self._add_port("yaw", NodeType.ELECTRICAL)
        self.yaw._add_node("i", NodeDirection.INPUT)
        self._add_port("pitch", NodeType.ELECTRICAL)
        self.pitch._add_node("i", NodeDirection.INPUT)

        self._register_node_coupling("SIGAMP_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("SIGPWR_P1o", self.pwr.i, self.p1.o)
        self._register_node_coupling("SIGPHS_P1o", self.phs.i, self.p1.o)
        self._register_node_coupling("SIGFRQ_P1o", self.frq.i, self.p1.o)

        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", NodeDirection.OUTPUT)
        self.mech._add_node("x", NodeDirection.OUTPUT)
        self.mech._add_node("y", NodeDirection.OUTPUT)
        self.mech._add_node("yaw", NodeDirection.OUTPUT)
        self.mech._add_node("pitch", NodeDirection.OUTPUT)

        self._register_node_coupling("dz_P1o", self.mech.z, self.p1.o)
        self._register_node_coupling("dx_P1o", self.mech.x, self.p1.o)
        self._register_node_coupling("dy_P1o", self.mech.y, self.p1.o)
        self._register_node_coupling("xbeta_P1o", self.mech.yaw, self.p1.o)
        self._register_node_coupling("ybeta_P1o", self.mech.pitch, self.p1.o)

        self.f = f
        self.P = P
        self.phase = phase
        self.__power_coeffs = self.DEFAULT_POWER_COEFFS.copy()
        self.add_gouy_phase = True
        self.signals_only = bool(signals_only)

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        self.dofs.pwr = LocalDegreeOfFreedom(
            f"{self.name}.dofs.pwr", self.P, self.pwr.i, 1
        )
        self.dofs.amp = LocalDegreeOfFreedom(
            f"{self.name}.dofs.amp", None, self.amp.i, 1
        )
        self.dofs.phs = LocalDegreeOfFreedom(
            f"{self.name}.dofs.phs", self.phase, self.phs.i, 1
        )
        self.dofs.frq = LocalDegreeOfFreedom(
            f"{self.name}.dofs.frq", self.f, self.frq.i, 1
        )
        self.dofs.z = LocalDegreeOfFreedom(f"{self.name}.dofs.z", None, self.mech.z, 1)
        self.dofs.x = LocalDegreeOfFreedom(f"{self.name}.dofs.x", None, self.mech.x, 1)
        self.dofs.y = LocalDegreeOfFreedom(f"{self.name}.dofs.y", None, self.mech.y, 1)
        self.dofs.yaw = LocalDegreeOfFreedom(
            f"{self.name}.dofs.yaw", None, self.mech.yaw, 1
        )
        self.dofs.pitch = LocalDegreeOfFreedom(
            f"{self.name}.dofs.pitch", None, self.mech.pitch, 1
        )

    def optical_equations(self):
        return {}

    def _source_frequencies(self):
        return [self.f.ref]

    def source_equation(self, node, f):
        """Returns optical carrier field to inject for a simulation."""
        E = self.get_output_field(self._model.homs)

        eps0_c = self._model._settings.EPSILON0_C
        if node is self.p1.o and (f is self.f.ref or f == self.f.value):
            scalar = np.sqrt(2 * self.P.ref / eps0_c) * np.exp(
                1.0j * np.pi / 180 * self.phase.ref
            )
            return scalar * E
        else:
            return None

    @property
    def power_coeffs(self):
        """The relative power factors and phase offsets for each HGnm mode.

        :`getter`: Returns the mode factors and phase offsets as a dict with the mode
                   indices as keys. Read-only.
        """
        return self.__power_coeffs.copy()

    def tem(self, n, m, factor, phase=0.0):
        """Distributes power into the mode HGnm.

        Parameters
        ----------
        n, m : int
            Mode indices.

        factor : float
            Relative power factor, modes with equal `factor` will
            have equivalent power distributed to them.

        phase : float, optional; default = 0.0
            Phase offset for the field, in degrees.

        Notes
        -----
        This does not change the total power of the laser, rather, it redistributes this
        power into / out of the specified mode.
        """
        self.__power_coeffs[(n, m)] = float(factor), float(phase)

    def __find_src_freq(self, sim):
        # If it's tunable we want to look for the symbol that is just this laser's
        # frequency, as it will be changing.
        for f in sim.optical_frequencies.frequencies:
            if not self.f.is_changing:
                # Don't match changing frequency bins if ours won't match.
                if not f.symbol.is_changing and (
                    f.f == self.f.value  # match potential param refs
                    or f.f == float(self.f.value)  # match numeric values
                ):
                    # If nothing is changing then we can just match freq values.
                    return f
            else:
                # If our frequency is changing then we have to have a frequency bin that
                # matches our symbol.
                if f.symbol == self.f.ref:
                    return f  # Simple case
        return None

    def _get_workspace(self, sim):
        from finesse.components.modal.laser import (
            laser_carrier_fill_rhs,
            laser_fill_signal,
            laser_fill_qnoise,
            LaserWorkspace,
            laser_set_gouy,
        )

        ws = LaserWorkspace(self, sim)
        ws.node_car_id = sim.carrier.node_id(self.p1.o)
        ws.fsrc_car_idx = -1

        ws.add_gouy_phase = bool(self.add_gouy_phase)
        ws.set_gouy_function(laser_set_gouy)
        # Carrier just fills RHS.
        ws.carrier.set_fill_rhs_fn(laser_carrier_fill_rhs)

        fsrc = self.__find_src_freq(sim.carrier)
        # Didn't find a Frequency bin for this laser in carrier simulation.
        if fsrc is None:
            raise Exception(f"Could not find a frequency bin at {self.f} for {self}")
        ws.fsrc_car_idx = fsrc.index

        if sim.is_modal:
            scaling = 0
            ws.power_coeffs = np.zeros(sim.model_settings.num_HOMs, dtype=np.complex128)
            coeffs = self.power_coeffs
            for i in range(sim.model_settings.num_HOMs):
                n = sim.model_settings.homs_view[i][0]
                m = sim.model_settings.homs_view[i][1]

                try:
                    factor, phase = coeffs.pop((n, m))
                except KeyError:
                    factor = phase = 0

                ws.power_coeffs[i] = crotate(
                    complex(math.sqrt(factor), 0), math.radians(phase)
                )
                scaling += abs(ws.power_coeffs[i]) ** 2

            if not scaling:
                raise RuntimeError(
                    f"No power in any modes of {self.name}! At least one mode must "
                    f"have a non-zero power factor applied to it."
                )

            for i in range(sim.model_settings.num_HOMs):
                ws.power_coeffs[i] /= np.sqrt(scaling)

            if coeffs:
                warn(
                    f"The following modes, included in the coeffs of "
                    f"{repr(self.name)}, are not being modelled and will be ignored: "
                    f"{list(coeffs.keys())}"
                )

        if sim.signal:
            ws.node_sig_id = sim.signal.node_id(self.p1.o)
            # Audio sim requies matrix filling
            # for signal couplings
            ws.signal.add_fill_function(
                laser_fill_signal, True
            )  # TODO sort out refill flag here
            ws.signal.set_fill_noise_function(NoiseType.QUANTUM, laser_fill_qnoise)
            # Find the sideband frequencies
            sb = tuple(
                (
                    f
                    for f in sim.signal.optical_frequencies.frequencies
                    if f.audio_carrier_index == fsrc.index
                )
            )
            if len(sb) != 2:
                raise Exception(
                    f"Only something other than two audio sidebands {sb} for carrier "
                    f"{fsrc}"
                )
            ws.fcar_sig_sb_idx = (sb[0].index, sb[1].index)

        # if sim.is_modal: self._update_tem_gouy_phases(sim)
        return ws

    def _couples_frequency(self, ws, connection, frequency_in, frequency_out):
        # The only connections we have are signal inputs to optical output
        # And all the inputs should generate any output.
        return True

    def set_output_field(self, E, homs):
        """Set optical field outputted using HOM vector.

        This changes the output power and mode content of the laser to match the
        requested ``E`` field.

        Parameters
        ----------
        E : sequence
            The complex optical field amplitude for `homs`.
        homs : sequence
            Sequence of (n, m) higher order modes. Typically this is just the
            ``model.homs`` value. It should match the size of ``E``.

        Notes
        -----
        If you do not want the gouy phase due to the current beam parameter values set
        at the laser to be added to the output, set the ``add_gouy_phase`` attribute of
        the laser element to ``False``.
        """
        E = np.asarray(E, dtype=complex)
        homs = np.array(homs, dtype=int)
        if homs.shape[1] != 2:
            raise ValueError("homs input should be a (N, 2) shape")
        if homs.shape[0] != E.shape[0]:
            raise ValueError("number of homs should match length of E field")
        if E.ndim != 1:
            raise ValueError("E field input is not a 1D array")

        self.power_coeffs.clear()
        self.P = sum(abs(E) ** 2)
        for (n, m), a in zip(homs, E):
            self.tem(n, m, abs(a) ** 2 / self.P.value, np.angle(a, deg=True))

    def get_output_field(self, homs):
        """Get optical field outputted as a HOM vector.

        Returns the complex amplitude of the modes specified in homs. This does not
        respect the `add_gouy_phase`` attribute of the laser element and will always
        return the complex amplitude without the gouy phase. If the gouy phase is
        required then it is recommended to use a FieldDetector at the laser output with
        ``add_gouy_phase`` set to ``True``.

        Parameters
        ----------
        homs : sequence
            Collection of (n, m) higher order modes to retrieve the output field for.
            Typically this is just the ``model.homs`` value. The output ``E`` vector
            will match the ordering of ``homs``.

        Returns
        -------
        sequence
            The output fields for `homs`. If a given (n, m) has no defined coefficients,
            its field defaults to 0.
        """
        E = np.zeros(len(homs), dtype=complex)
        total_power = self.P.value
        for i, (n, m) in enumerate(homs):
            power_frac, phase = self.power_coeffs.get((n, m), (0, 0))
            E[i] = np.sqrt(total_power * power_frac) * np.exp(1j * phase / 180 * np.pi)
        return E



"""Base class for dependencies of beam tracing routines."""

from numbers import Number

from finesse.element import ModelElement


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class TraceDependency(ModelElement):
    """Base for classes which are dependency objects of beam tracing.

    Both :class:`.Cavity` and :class:`.Gauss` inherit from this class.

    Parameters
    ----------
    name : str
        Name of newly created trace dependency.

    priority : number, optional; default: 0
        Priority value for beam tracing. See :attr:`.TraceDependency.priority`.
    """

    def __init__(self, name, priority=0):
        super().__init__(name)

        self.__priority = _check_priority(priority)

    @property
    def priority(self):
        """The priority value of the dependency.

        Larger values indicate higher priority for beam tracing. The list
        :attr:`.Model.trace_order` is sorted in descending order of trace
        dependency priority values. Dependencies with equal priority values
        will be ordered alphabetically by their names.

        :`getter`: Returns the priority value.
        :`setter`: Sets the priority value and re-sorts the associated
                   :attr:`.Model.trace_order` accordingly.
        """
        return self.__priority

    @priority.setter
    def priority(self, value):
        self.__priority = _check_priority(value)
        self._model._resort_trace_dependencies()

    def take_priority(self):
        """Give this trace dependency the highest priority in the model.

        Internally this will set the priority of this dependency to the current highest
        priority in the model plus one.
        """
        self.priority = 1 + self._model.trace_order[0].priority

    def make_final(self):
        """Give this trace dependency the lowest priority in the model.

        Internally this will set the priority of this dependency to the current lowest
        priority in the model minus one.
        """
        self.priority = self._model.trace_order[-1].priority - 1


def _check_priority(value):
    if not isinstance(value, Number):
        raise TypeError("Expected priority value to be a number.")

    return value



import typing


def issubtype(cls, annotation):
    """Check that `cls` is compatible with typing annotation `annotation`.

    Parameters
    ----------
    cls : type
        The type to check for membership of `annotation`.

    annotation : :mod:`typing` type
        The annotation to check.
    """
    if typing.get_origin(annotation) is None:
        # This is not a typing type, but a normal Python type.
        return annotation is cls
    # Maybe a generic.
    return any(issubtype(cls, arg) for arg in typing.get_args(annotation))



# -*- coding: utf-8 -*-
"""Holds the various instances of simulation classes.

Listed below are all the sub-modules of the ``simulations`` module with a brief
description of the contents of each.
"""

from .simulation import BaseSimulation
from .basesolver import BaseSolver

# from .sparse.simulation import SparseMatrixSimulation

# from finesse.simulations.digraph import DigraphSimulation, DigraphSimulationBase
# from finesse.simulations.debug import DebugSimulation
# from finesse.simulations.dense import DenseSimulation

__all__ = (
    # "AccessSimulation",
    "BaseSimulation",
    "BaseSolver",
    # "SparseMatrixSimulation",
    # "DigraphSimulation",
    # "DigraphSimulationBase",
    # "DebugSimulation",
    # "DenseSimulation",
)



import numpy as np

from ...solutions import BaseSolution
from .base import Action
from ...components.node import OpticalNode


class DCFieldsSolutions(BaseSolution):
    """Contains the result of a :class:`DCFields` action. The returned fields are the
    internal description of the carrier are all nodes, modes, and frequencies. Fields
    are defined as P=|E|^2.

    Attributes
    ----------
    homs : list
        The Hermite-Gaussian higher order mode indices at each node, of each frequency.
    nodes : tuple[str]
        Name of nodes the fields were extracted at
    frequencies : array_like[dtype=float]
        Values of the optical frequencies at each node
    fields : array_like
        Field data array with shape [nodes, frequencies, HOMs.shape[0]]

    Notes
    -----
    The `fields` attribute will contain all the required information with its ordering
    defined by the `nodes`, `frequencies`, and `homs` array. For easier indexing you can
    access the solution directly with node names.

    Select all the frequencies and modes at l1.p1.i:
    >>> sol['l1.p1.i']

    Select all the first modes at all frequencies and modes at l1.p1.i:
    >>> sol['l1.p1.i', :, 0]

    Select all the frequencies and modes at two nodes:
    >>> sol[('l1.p1.i', 'l1.p1.o')]

    The same nodes can be selected directly with the OpticalNodes:
    >>> sol[(model.l1.p1.i, model.l1.p1.o)]

    Select all the HG02 modes at every node and frequency
    >>> sol[:, :, sol.homs.index([0, 2]))]

    Select all the HG02 modes at every output node and every frequency
    >>> idx = np.array([node.split(".")[-1] == "o" for node in sol.nodes])
    >>> sol[idx, :, sol.homs.index([0, 2])]

    The frequency and mode selection must be a slice or an array of integer or boolean
    values.
    """

    homs = None
    nodes = None
    frequencies = None
    fields = None

    def selector(self, key):
        def node_name(key):
            if isinstance(key, str):
                return key
            elif isinstance(key, OpticalNode):
                return key.full_name
            else:
                raise TypeError("key should be a string or optical node")

        if isinstance(key, (str, int, slice, np.ndarray, OpticalNode)):
            # specifying a single node or a slice of multiple nodes
            key = [key]
        elif isinstance(key, tuple):
            # this is either a tuple specifying multiple nodes or
            # specifying multiple dimensions
            assert len(key) > 1
            if len(key) in [2, 3]:
                if isinstance(key[1], (str, OpticalNode)):
                    # specifying multiple nodes
                    key = [key]
                else:
                    # specifying multiple dimensions
                    key = list(key)
            else:
                # specifying multiple nodes
                key = [list(key)]
        else:
            raise TypeError(
                f"Unrecognized key type {type(key)}."
                "Multiple nodes should be specified by tuples, slices, or ndarrays."
            )
        if not isinstance(key[0], (slice, int, np.ndarray)):
            nodes = np.array([node_name(n) for n in np.atleast_1d(key[0])])
            for i in nodes:
                if i not in self.nodes:
                    raise KeyError(f"`{i}` is not in the solution nodes")
            key[0] = tuple(self.nodes.index(i) for i in nodes)

        return tuple(key)

    def __getitem__(self, key):
        if isinstance(key, str) and self.name == key:
            return self
        else:
            return self.fields[self.selector(key)]


class DCFields(Action):
    """An action that saves the DC (carrier) fields at all nodes, frequencies, and
    higher order modes for the current state of the simulation.

    Parameters
    ----------
    name : str, optional
        Name of the solution generated by this action
    """

    def __init__(self, *, name="dcfields"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        sol = DCFieldsSolutions(self.name)
        # do the DC solve again just in case a parameter has changed and another
        # action hasn't done so already
        state.sim.run_carrier()
        sol.homs = state.model.homs.tolist()
        sol.nodes = tuple(state.sim.carrier.nodes.keys())
        sol.frequencies = np.array(
            [f.f for f in state.sim.carrier.optical_frequencies.frequencies]
        )
        sol.fields = np.reshape(
            state.sim.carrier.M().rhs_view[0, :].copy(),
            (len(sol.nodes), len(sol.frequencies), len(sol.homs)),
            order="C",
        )
        # rescale into sqrt(W) units
        sol.fields /= np.sqrt(2)
        return sol



from __future__ import annotations

import enum
import inspect
import json
import platform
import re
from io import StringIO
import importlib.metadata

try:
    import readline

    readline_available = True
    # probably Mac
except ImportError:
    readline_available = False
import sys
import traceback
import warnings
from pathlib import Path
from types import ModuleType

import finesse
from finesse import is_interactive

# General note for this entire file: multiline strings that have zero indentation
# cause issues with the docs build. Please don't change the ugly multiline strings
# without checking!


class SourceType(enum.Enum):
    INTERACTIVE = "Interactive"
    SCRIPT = "Script"
    REPL = "REPL"
    STDIN = "stdin"


def get_package_versions() -> str:
    """Report all currently imported package version by looping over :attr:`sys.modules`
    and looking for '__version__' attributes. Explicitly avoids calling into conda/pip
    since there are too many package managers to accommodate for.

    Returns
    -------
    str
        list of <package> == <version> for every package
    """
    modules = []
    unversioned_modules = []
    for mod in sys.modules.values():
        if isinstance(mod, ModuleType):
            if not any(char in mod.__name__ for char in (".", "/")):
                try:
                    version = importlib.metadata.version(mod.__name__)
                    modules.append(f"{mod.__name__} == {version}")
                except importlib.metadata.PackageNotFoundError:
                    unversioned_modules.append(f"{mod.__name__}")
    modules.sort()
    unversioned_modules.sort()
    return "\n".join(modules) + "\n" + "\n".join(unversioned_modules) + "\n"


def get_source() -> str:
    """Get source of the '__main__' module. Supports Ipython (Jupyter Notebook, VSCode),
    interactive interpreter and regular python modules.

    Returns
    -------
    str
        Source code
    """
    main = sys.modules["__main__"]
    source_type = get_source_type()
    if source_type == SourceType.INTERACTIVE:
        # Undocumented module attributes that might store the file path
        for file_attr in ("__file__", "__vsc_ipynb_file__", "__session__"):
            if source_fn := getattr(main, file_attr, False):
                source_fn = Path(str(source_fn))
                if source_fn.suffix == ".ipynb":
                    with open(source_fn, "r") as f:
                        return ipynb_to_md(json.load(f))
        # otherwise concatenate lines from code cells (without markdown cells)
        source = "\n".join(main.In)
    # interactive interpreter: no distinction in history file between different sessions
    elif source_type == SourceType.REPL:
        warnings.warn(
            "Using last 20 commands to generate bug report from interactive interpreter",
            stacklevel=1,
        )
        if not readline_available:
            warnings.warn(
                "Can not read REPL history!",
                stacklevel=1,
            )
            return ""
        source = ""
        hist_length = readline.get_current_history_length()
        n_lines = min(hist_length, 20)
        for i in range(hist_length - n_lines, hist_length):
            source += str(readline.get_history_item(i)) + "\n"
    # normal .py file
    elif source_type == SourceType.SCRIPT:
        source = inspect.getsource(main)
    elif source_type == SourceType.STDIN:
        # Should maybe be an exception, but we don't really want to raise exceptions
        # in code meant to handle exceptions
        warnings.warn(
            RuntimeWarning("Can not get source when passing python code via stdin"),
            stacklevel=2,
        )
        source = ""
    else:
        raise ValueError(f"Unknown source type {source_type}")
    return source.strip()


def get_source_type() -> SourceType:
    """Type of source for the python code currently being executed.

    Returns
    -------
    SourceType
        Interactive environment (jupyter), terminal REPL or plain python script
    """
    if is_interactive():
        return SourceType.INTERACTIVE
    elif not hasattr(sys.modules["__main__"], "__file__"):
        return SourceType.REPL
    elif sys.modules["__main__"].__file__ == sys.stdin.name:
        return SourceType.STDIN
    else:
        return SourceType.SCRIPT


def ipynb_to_md(ipynb: dict) -> str:
    """Converts notebook json object to markdown. Extracts markdown cells as raw text
    and code blocks wrapped in a python code block.

    Parameters
    ----------
    ipynb : dict
        notebook json dict

    Returns
    -------
    str
        Markdown representing notebook
    """
    md = ""
    for cell in ipynb["cells"]:
        if cell["cell_type"] == "code":
            lang = "python"
        elif cell["cell_type"] == "markdown":
            lang = "markdown"
        source = "".join(cell["source"])
        md += wrap_block(source, lang=lang)

    return md


def wrap_block(code: str, lang: str = "python") -> str:
    """Wraps a string in a markdown code block like.

    ```python
    print('foo')
    ```

    Parameters
    ----------
    code : str
        code to wrap
    lang : str, optional
        language of code, by default "python"

    Returns
    -------
    str
        Markdown code block
    """
    return f"```{lang}\n{code}\n```\n"


def get_formatted_traceback() -> str:
    if sys.version_info.minor < 12:
        try:
            io = StringIO()
            traceback.print_last(file=io)
            return io.getvalue()
        except ValueError:
            pass
    if exc := getattr(sys, "last_exc", False):
        assert isinstance(exc, BaseException)
        return "\n".join(traceback.format_exception(exc))
    else:
        return traceback.format_exc()


def get_formatted_argv() -> str:
    args = [f"`{arg}`" for arg in sys.argv if len(arg)]
    if len(args):
        return "### Arguments\n\n" + " ".join(args)
    else:
        return ""


def get_formatted_source() -> str:
    source = get_source()
    if len(source) > 1:
        return (
            f"### Source [{get_source_type()}]\n" "\n" "Showing last 20 lines\n"
            if get_source_type() == SourceType.REPL
            else ""
            f"{source if get_source_type() == SourceType.INTERACTIVE else wrap_block(source)}\n"
        )
    else:
        return ""


def bug_report(
    title: str | None = None,
    file: str | Path | None = None,
    include_source: bool = False,
):
    """Generate a markdown bug report, suitable for copy-pasting into chatrooms or
    GitLab issues. Contains the source code, the triggered exception (if any) and
    machine and python environment information.

    Parameters
    ----------
    title : str | None, optional
        Title to insert on top of markdown, by default None
    file : str | Path | None, optional
        Whether to write the report to file. Will silently overwrite existing files,
        by default None
    include_source : bool, optional
        Wether to include the source code that caused the exception (the contents of the
        Jupyter notebook or Python script file) into the bug report. Be careful when
        including source with proprietary/confidential information source in bug reports
        shared in public spaces like Gitlab or the Matrix channel. Defaults to False
    """
    # 4 spaces is equal to a code block in gitlab markdown!
    report = (
        f"# {title if title else 'Finesse3 bug report'}\n"
        "\n"
        "## Environment\n"
        "\n"
        f"- **Finesse version:** `{finesse.__version__}`\n"
        f"- **Python version:** `{sys.version}`\n"
        f"- **Platform:** `{platform.system()} {platform.machine()}`\n"
        "\n"
        "## Entry point\n"
        "\n"
        f"`{sys.executable}`\n"
        "\n"
        f"{get_formatted_argv()}\n"
        "\n"
        f"{get_formatted_source() if include_source else ''}\n"
        "\n"
        "## Stack trace\n"
        "\n" + wrap_block(get_formatted_traceback(), lang="text") + "\n"
        "## Package versions\n"
        "\n" + wrap_block(get_package_versions(), lang="text")
    )
    # remove excessive empty lines due to removed sections
    report = re.sub(r"\n{2,}", repl=r"\n\n", string=report)
    if file:
        file = Path(file)
        file.write_text(report)
        print(f"Bug report written to {Path.cwd() / file}\n")
    return report



from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.element_workspace cimport BaseCValues
from finesse.cymath.cmatrix cimport SubCCSView, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.simulations.base cimport ModelSettings
from finesse.frequency cimport frequency_info_t
from finesse.components.modal.workspace cimport KnmConnectorWorkspace
from finesse.cymath cimport complex_t
from cpython.ref cimport PyObject
from numpy cimport ndarray
import numpy as np
cimport numpy as np
from finesse.enums cimport ModulatorType


cdef struct modulator_coupling_order:
    Py_ssize_t f1_index
    Py_ssize_t f2_index
    Py_ssize_t order_index
    int order


cdef class ModulatorValues(BaseCValues):
    cdef public:
        double f
        double midx
        double phase
        int mod_type
        int order
        bint positive_only

cdef struct modulator_optical_connections:
    # Couples frequencies
    PyObject*** P1i_P2o
    PyObject*** P2i_P1o

cdef struct modulator_signal_connections:
    # Couples frequencies
    PyObject*** amp_P1o
    PyObject*** amp_P2o
    PyObject*** phs_P1o
    PyObject*** phs_P2o

cdef class ModulatorOpticalConnections:
    cdef public:
        int P1i_P2o_idx
        int P2i_P1o_idx

    cdef readonly:
        SubCCSView2DArray P1i_P2o
        SubCCSView2DArray P2i_P1o
    cdef:
        modulator_optical_connections opt_conn_ptrs


cdef class ModulatorSignalConnections(ModulatorOpticalConnections):
    cdef public:
        int amp_P1o_idx
        int phs_P1o_idx
        int amp_P2o_idx
        int phs_P2o_idx

    cdef readonly:
        SubCCSView2DArray amp_P1o
        SubCCSView2DArray amp_P2o
        SubCCSView2DArray phs_P1o
        SubCCSView2DArray phs_P2o
    cdef:
        modulator_signal_connections sig_conn_ptrs


cdef class ModulatorWorkspace(KnmConnectorWorkspace):
    cdef public:
        # Complete scattering matrices for each propagation direction
        KnmMatrix K12
        KnmMatrix K21

        # Arrays of scattering losses for each mode coupling
        double[::1] K12_loss
        double[::1] K21_loss

        # Refractive indices of adjacent spaces
        double nr1
        double nr2

    cdef readonly:
        dict carrier_frequency_couplings
        dict signal_frequency_couplings
        bint amp_signal_enabled
        bint phs_signal_enabled
    cdef:
        int N_orders
        ndarray orders
        ndarray amps
        ndarray phases
        ndarray factors_12
        ndarray factors_21
        modulator_coupling_order* sig_coupling_orders
        modulator_coupling_order* car_coupling_orders
        Py_ssize_t N_sig_coupling_orders
        Py_ssize_t N_car_coupling_orders
        bint* increment_sig
        ModulatorValues cvalues
        ModulatorOpticalConnections mcc
        ModulatorSignalConnections mcs
        complex_t[:, ::1] eye  # nhom x nhom diagonal matrix for noise coupling filling

    cpdef fill_quantum_matrix(self)



import pytest
import finesse
import numpy as np
import warnings
from finesse.warnings import CavityUnstableWarning


@pytest.fixture
def coupled_cavity():
    model = finesse.Model()
    model.parse(
        """
    laser l1 P=40
    m m1 T=0.04835 L=30u Rc=-1430.0 phi=90
    s s1 m1.p2 m2.p1 L=5.9399
    m m2 T=0.01377 L=27u Rc=-1424.6
    s s2 m2.p2 m3.p1 L=2999.8
    m m3 T=4.4u L=27u Rc=1790.0
    link(l1, m1)
    cav cav1 m1.p2.o
    cav cav2 m2.p2.o
    modes(even, maxtem=2)
    fd E1 m2.p1.o 0
    fd E2 m2.p2.o 0
    """
    )
    return model


@pytest.mark.parametrize(
    "param",
    [
        "m1.Rcx",
        "m1.Rcy",
        "m2.Rcx",
        "m2.Rcy",
        "m3.Rcx",
        "m3.Rcy",
        "s1.L",
        "s2.L",
        "m1.phi",
        "m2.phi",
        "m3.phi",
    ],
)
def test_change_param(coupled_cavity, param):
    """Ensure that changing some geometric parameters in a coupled cavity actually
    recomputes and solves correctly.

    Just checks that the numbers are actually different. during the axis sweep.
    """
    model = coupled_cavity.deepcopy()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=CavityUnstableWarning)
        sol = model.run(f"xaxis({param}, lin, 0, 1, 1, True)")
    assert not np.allclose(sol["E1"][0], sol["E1"][1])
    assert not np.allclose(sol["E2"][0], sol["E2"][1])



from finesse.tracing.tree cimport TraceTree, DependencyType
from finesse.utilities.collections cimport OrderedSet


cdef class tree_container:
    cdef:
        list trees
        Py_ssize_t size

        # All the optical nodes covered by all the trees
        OrderedSet node_coverage

    cdef clear(self)
    cdef append(self, TraceTree tree)
    cdef remove(self, TraceTree tree)
    cdef _update_after_sub_remove(self, TraceTree sub_tree)


cdef class TraceForest:
    cdef public:
        bint symmetric # Should beam traces on this forest set opposite node q to -q*?

    cdef readonly:
        tree_container forest

        # The list of TraceDependency objects in order of tracing priority
        list dependencies

        object model

    ### Planting and clearing ###

    cpdef void clear(self) noexcept
    cdef void trim(self) noexcept

    cdef int _add_internal_cavity_trees(self, list cavities) except -1
    cdef int _handle_overlapping_cavities(self) except -1
    cdef int _add_external_cavity_tree(self, object cav) except -1
    cdef int _add_gauss_tree(self, object gauss) except -1

    cdef int _add_beamsplitter_branch_trees(self) except -1
    cdef int _add_backwards_nonsymm_trees(self) except -1

    cpdef OrderedSet find_untraversed_nodes(self)

    ### Searching and attributes ###

    cpdef Py_ssize_t size(self) noexcept
    cpdef bint empty(self) noexcept

    cpdef bint contains(self, object o) noexcept

    # Get a list of trees based on a predicate
    cdef list _get_trees_upred(self, bint (*predicate)(TraceTree))
    cdef list _get_trees_bpred(self, bint (*predicate)(TraceTree, object), object o)

    cpdef list trees_of_dependency(self, object dependency)
    cdef list get_internal_cavity_trees(self)
    cdef list get_external_cavity_trees(self)
    cdef list get_gauss_trees(self)

    cpdef TraceTree find_tree_from_node(self, object node)
    cpdef object find_dependency_from_node(self, object node, bint raise_not_found=?)

    cdef object node_from_name(self, name)

    ### Changing geometric parameter forest algorithms ###

    cpdef TraceForest make_changing_forest(self)

    ### Automatic mode mismatch coupling determination ###

    cpdef tuple find_potential_mismatch_couplings(self, TraceForest other=?)
    cpdef tuple find_intersection_couplings(self, TraceForest other=?)
    cpdef tuple get_mirror_reflection_couplings(
        self,
        bint ignore_internal_cavities=?,
        list skip_dependencies=?,
    )

    ### Drawing ###

    cpdef draw_by_dependency(self)
    cpdef draw(self)

    ### Propagating beams ###

    cpdef dict trace_beam(self)



# %%
import pytest
import numpy as np
import finesse
from finesse import Model
from finesse.analysis.actions import (
    Xaxis,
    FrequencyResponse,
    FrequencyResponse2,
    FrequencyResponse3,
    FrequencyResponse4,
)
from finesse.components import (
    Squeezer,
    Beamsplitter,
    Laser,
    ReadoutDC,
)
import finesse.components as fc
import finesse.analysis.actions as fa


@pytest.fixture
def squeezer():
    model = finesse.Model()
    model.add(Squeezer("SQZ", 0))
    model.add(Beamsplitter("BS", T=1e-14, L=0))
    model.add(Laser("L0", P=1e14, phase=-90))
    model.add(ReadoutDC("PD", model.BS.p2.o))
    model.connect(model.SQZ.p1, model.BS.p1, 1000)
    model.connect(model.L0.p1, model.BS.p4)
    model.fsig.f = 1
    return model


def _assert_sqz_response_phase(u2u, l2l, u2l=None, l2u=None):
    # checks that the upper and conjugate lower sidebands have the same phase unity amplitude
    # and that if upper to lower and lower to upper are calculated, they are zero
    assert np.allclose(u2u, l2l)
    for tf in [u2u, l2l]:
        assert np.allclose(np.abs(tf), 1)
    for tf in [u2l, l2u]:
        if tf is not None:
            assert np.allclose(np.abs(tf), 0)


@pytest.mark.skip(reason="needs more thought on factor of two")
def test_sqz_sideband_frequency_response_phase(squeezer):
    model = squeezer

    sol = model.run(
        FrequencyResponse(
            np.geomspace(0.1, 1000, 3),
            [model.SQZ.upper, model.SQZ.lower_conj],
            model.PD.DC.o,
        )
    )
    _assert_sqz_response_phase(
        u2u=sol["PD.DC.o", "SQZ.upper"],
        l2l=sol["PD.DC.o", "SQZ.lower_conj"],
    )


@pytest.mark.skip(reason="needs more thought on factor of two")
def test_sqz_sideband_frequency_response_2_phase(squeezer):
    model = squeezer

    sol = model.run(
        FrequencyResponse2(
            np.geomspace(0.1, 1000),
            [(model.SQZ.p1.o, model.fsig.f.ref), (model.SQZ.p1.o, -model.fsig.f.ref)],
            [model.PD.DC.o],
        )
    )
    upper = sol.out[:, 0, 0, 0]
    lower_conj = sol.out[:, 0, 1, 0]
    _assert_sqz_response_phase(u2u=upper, l2l=lower_conj)


def test_sqz_sideband_frequency_response_3_phase(squeezer):
    model = squeezer
    f_u = +model.fsig.f.ref
    f_l = -model.fsig.f.ref

    sol = model.run(
        FrequencyResponse3(
            np.geomspace(0.1, 1000, 10),
            [(model.SQZ.p1.o, f_u), (model.SQZ.p1.o, f_l)],
            [(model.PD.p1.i, f_u), (model.PD.p1.i, f_l)],
        )
    )
    u2u = sol.out[:, 0, 0, 0, 0]
    u2l = sol.out[:, 1, 0, 0, 0]
    l2u = sol.out[:, 0, 1, 0, 0]
    l2l = sol.out[:, 1, 1, 0, 0]
    _assert_sqz_response_phase(u2u=u2u, u2l=u2l, l2u=l2u, l2l=l2l)


def test_sqz_sideband_frequency_response_4_phase(squeezer):
    model = squeezer
    fsig = model.fsig.f.ref

    sol = model.run(
        FrequencyResponse4(
            np.geomspace(0.1, 1000, 10),
            [model.SQZ.upper, model.SQZ.lower_conj],
            [
                (model.BS.p1.i, +fsig),
                (model.BS.p1.i, -fsig),
            ],
        )
    )
    u2u = sol.out[:, 0, 0, 0]
    u2l = sol.out[:, 1, 0, 0]
    l2u = sol.out[:, 0, 1, 0]
    l2l = sol.out[:, 1, 1, 0]
    _assert_sqz_response_phase(u2u=u2u, u2l=u2l, l2u=l2u, l2l=l2l)


@pytest.mark.parametrize("f", [1, "fsig.f"])
def test_frequency_response2_exception(f, squeezer):
    model = squeezer
    if isinstance(f, str):
        f_u = "+" + f
        f_l = "-" + f
    else:
        f_u = +f
        f_l = -f

    action = FrequencyResponse2(
        np.geomspace(0.1, 1000, 1),
        [(model.SQZ.p1.o, f_u), (model.SQZ.p1.o, f_l)],
        ["PD.DC"],
    )

    with pytest.raises(finesse.exceptions.FinesseException):
        model.run(action)


@pytest.mark.parametrize("f", [1, "fsig.f"])
def test_frequency_response3_exception(f, squeezer):
    model = squeezer
    if isinstance(f, str):
        f_u = "+" + f
        f_l = "-" + f
    else:
        f_u = +f
        f_l = -f

    action = FrequencyResponse3(
        np.geomspace(0.1, 1000, 1),
        [(model.SQZ.p1.o, f_u), (model.SQZ.p1.o, f_l)],
        [(model.PD.p1.i, f_u), (model.PD.p1.i, f_l)],
    )

    with pytest.raises(finesse.exceptions.FinesseException):
        model.run(action)


@pytest.mark.parametrize("f", [1, "fsig.f"])
def test_frequency_response4_exception(f, squeezer):
    model = squeezer
    if isinstance(f, str):
        f_u = "+" + f
        f_l = "-" + f
    else:
        f_u = +f
        f_l = -f

    action = FrequencyResponse4(
        np.geomspace(0.1, 1000, 1),
        [model.SQZ.upper, model.SQZ.lower_conj],
        [(model.PD.p1.i, f_u), (model.PD.p1.i, f_l)],
    )

    with pytest.raises(finesse.exceptions.FinesseException):
        model.run(action)


def test_frequency_response3_diff_inp_out():
    model = Model()
    model.add(fc.Mirror("ETM", T=0, L=0))
    model.add(fc.Mirror("ITM", T=0.014, L=0))
    model.connect(model.ITM.p1, model.ETM.p1, L=4e3)
    model.fsig.f = 1
    F_Hz = np.geomspace(1, 5e3, 10)
    fsig = model.fsig.f.ref
    model.run(
        fa.FrequencyResponse3(
            F_Hz,
            [
                (model.ITM.p2.i, +fsig),
                (model.ITM.p2.i, -fsig),
                (model.ETM.p1.o, +fsig),
                (model.ETM.p1.o, -fsig),
            ],
            [
                (model.ITM.p2.o, +fsig),
                (model.ITM.p2.o, -fsig),
            ],
        )
    )


@pytest.mark.parametrize("F_Hz", [[1], [1, 2], [1, 2, 3]])
def test_output_shape(F_Hz):
    F_Hz = np.asarray(F_Hz)
    from finesse import Model
    import finesse.analysis.actions as act
    from finesse.components import Squeezer, Mirror, Cavity, ReadoutDC

    model = Model()
    model.add(Squeezer("SQZ", db=10, angle=0))
    model.add(Mirror("ITM", T=0.014, L=0, Rc=1935))
    model.add(Mirror("ETM", T=0, L=0, Rc=2245))
    model.connect(model.ITM.p1, model.ETM.p1, 4e3)
    model.connect(model.SQZ.p1, model.ITM.p2)
    model.add(Cavity("cavARM", model.ETM.p1.o))
    model.add(ReadoutDC("REFL", model.ETM.p1.o, output_detectors=True))
    model.modes(maxtem=2)
    model.fsig.f = 1

    def run_fresp_actions(F_Hz):
        fsig = model.fsig.f.ref
        sol = model.run(
            act.Series(
                act.FrequencyResponse(
                    F_Hz,
                    ["SQZ.upper", "SQZ.lower_conj"],
                    ["REFL.DC.o"],
                    name="fresp1",
                ),
                act.FrequencyResponse2(
                    F_Hz,
                    [
                        ("SQZ.p1", +fsig),
                        ("SQZ.p1", -fsig),
                    ],
                    ["REFL.DC.o"],
                    name="fresp2",
                ),
                act.FrequencyResponse3(
                    F_Hz,
                    [
                        ("SQZ.p1.o", +fsig),
                        ("SQZ.p1.o", -fsig),
                    ],
                    [
                        ("ETM.p1.i", +fsig),
                        ("ETM.p1.i", -fsig),
                    ],
                    name="fresp3",
                ),
                act.FrequencyResponse4(
                    F_Hz,
                    ["ETM.mech.z"],
                    [
                        ("ITM.p2.o", +fsig),
                        ("ITM.p2.o", -fsig),
                    ],
                    name="fresp4",
                ),
            )
        )
        return sol

    sol = run_fresp_actions(F_Hz)
    assert sol["fresp1"].out.shape == (F_Hz.size, 1, 2)  # outputs, inputs
    assert sol["fresp2"].out.shape == (F_Hz.size, 1, 2, 6)  # outputs, inputs, homs
    assert sol["fresp3"].out.shape == (
        F_Hz.size,
        2,
        2,
        6,
        6,
    )  # outputs, inputs, homs, homs
    assert sol["fresp4"].out.shape == (F_Hz.size, 2, 1, 6)  # outputs, inputs, homs


def test_FrequencyResponseSolution_getattr_inverse():
    """Tests if [inputs, outputs] also works until it's deprecated."""
    from finesse.analysis.actions.lti import FrequencyResponseSolution

    t = FrequencyResponseSolution("name")
    t.inputs = ("A", "B", "C")
    t.outputs = ("D", "E", "F", "G")
    t.out = np.random.rand(3, len(t.outputs), len(t.inputs))

    with pytest.warns(DeprecationWarning):
        assert np.all(t[("F", "G"), ("C", "A")] == t[("C", "A"), ("F", "G")])


def test_FrequencyResponseSolution_getattr():
    from finesse.analysis.actions.lti import FrequencyResponseSolution

    t = FrequencyResponseSolution("name")
    t.inputs = ("A", "B", "C")
    t.outputs = ("D", "E", "F", "G")
    t.out = np.random.rand(3, len(t.outputs), len(t.inputs))

    assert t["name"] is t, "Didn't return the solution itself"

    with pytest.raises(KeyError):
        _ = t[1]

    with pytest.raises(KeyError):
        _ = t["1"]

    with pytest.raises(KeyError):
        _ = t[b"1"]

    with pytest.raises(KeyError):
        _ = t[1, 2, 3]

    with pytest.raises(KeyError):
        _ = t[:]

    assert np.all(t["D", "A"] == t.out[:, 0, 0])
    assert np.all(t["D", "C"] == t.out[:, 0, 2])
    assert np.all(t["F", "C"] == t.out[:, 2, 2])
    assert np.all(t["F", ("C", "A")] == t.out[:, 2, [2, 0]])
    assert np.all(t[("F", "G"), ("C", "A")] == t.out[:, [2, 3], [2, 0]])
    assert np.all(t["D", :] == t.out[:, 0, :])
    assert np.all(t["D", ::2] == t.out[:, 0, ::2])
    assert np.all(t["D", 1:] == t.out[:, 0, 1:])
    assert np.all(t[:, "B"] == t.out[:, :, 1])
    assert np.all(t[::2, "B"] == t.out[:, ::2, 1])
    assert np.all(t[1:, "B"] == t.out[:, 1:, 1])
    assert np.all(t[:, :] == t.out)

    with pytest.raises(KeyError):
        _ = t["Z", :]
    with pytest.raises(KeyError):
        _ = t[:, "Z"]
    with pytest.raises(KeyError):
        _ = t["D", "Z"]  # non existent input
    with pytest.raises(KeyError):
        _ = t["Z", "A"]  # non existent output


@pytest.fixture
def model():
    model = Model()
    model.add(fc.Laser("L", P=10))
    model.add(fc.ReadoutDC("PD"))
    model.link("L", "PD")
    model.fsig.f = 1
    return model


def test_frequency_response2_laser_pd(model):
    F_Hz = [10]
    fsig = model.fsig.f.ref
    DC = model.run("dc_fields()")
    sol2 = model.run(
        fa.FrequencyResponse2(
            F_Hz,
            [
                (model.L.p1.o, +fsig),
                (model.L.p1.o, -fsig),
            ],
            [
                model.PD.DC.o,
            ],
        )
    )

    Ec = DC["PD.p1.i"].squeeze()

    assert np.allclose(  # Half a carrier modulation in each sideband
        sol2.out.squeeze() @ (np.array([0.25, 0.25]) * Ec), 10
    )
    # Single sideband modulation
    assert np.allclose(sol2.out.squeeze() @ (np.array([0.5, 0]) * Ec), 10, atol=1e-14)
    assert np.allclose(sol2.out.squeeze() @ (np.array([0, 0.5]) * Ec), 10, atol=1e-14)


def test_frequency_response_laser_pd_product(model):
    F_Hz = [10]
    fsig = model.fsig.f.ref
    sol1 = model.run(
        fa.FrequencyResponse(
            F_Hz,
            [
                model.L.amp,
            ],
            [
                model.PD.DC.o,
            ],
        )
    )

    sol2 = model.run(
        fa.FrequencyResponse2(
            F_Hz,
            [
                (model.L.p1.o, +fsig),
                (model.L.p1.o, -fsig),
            ],
            [
                model.PD.DC.o,
            ],
        )
    )

    sol4 = model.run(
        fa.FrequencyResponse4(
            F_Hz,
            [
                model.L.amp,
            ],
            [
                (model.L.p1.o, +fsig),
                (model.L.p1.o, -fsig),
            ],
        )
    )
    # Should be the equivalent
    assert np.allclose(
        sol4.out.squeeze() @ sol2.out.squeeze(), sol1.out.squeeze(), atol=1e-14
    )


@pytest.mark.parametrize("type", ["amp", "phs", "pwr", "frq"])
def test_laser_siggen_equals_frequency_response_2(type):
    model = finesse.script.parse(
        f"""
        fsig(1)
        l l1 P=1
        readout_dc PD
        link(l1, PD)
        ad up l1.p1.o +fsig
        ad lo l1.p1.o -fsig
        sgen sg l1.{type}.i amplitude=1 phase=0
        """
    )
    out = model.run()
    sol = model.run(
        f"frequency_response4([1], [l1.{type}.i], [[l1.p1.o, +fsig], [l1.p1.o, -fsig]])"
    )
    H = sol.out.squeeze()
    assert np.allclose(H[0], out["up"], atol=1e-14)
    assert np.allclose(H[1].conj(), out["lo"], atol=1e-14)


def test_frequency_response_equals_sgen():
    f_min = 1e3
    f_max = 1e8
    steps = 100
    F = np.logspace(np.log10(f_min), np.log10(f_max), steps + 1)
    base_model = finesse.Model()
    base_model.parse(
        """
    l l1 P=1

    s sp1 l1.p1 eom.p1 L=0
    mod eom 15M 0.7 order=1
    s sp2 eom.p2 m1.p1 L=0

    m m1 R=0.9 L=0
    s cav m1.p2 m2.p1 L=1
    m m2 R=0.9 L=0
    """
    )
    readout_model = base_model.deepcopy()
    readout_model.parse(
        """
    readout_rf readout m1.p1.o f=eom.f phase=0
    fsig(1)
    """
    )
    readout_sol = readout_model.run(
        FrequencyResponse(
            F,
            inputs=[readout_model.m2.mech.z],
            outputs=[readout_model.readout.I],
        )
    )
    readout_tf = readout_sol["frequency_response"]["readout.I", "m2.mech.z"]

    pd2_model = base_model.deepcopy()
    pd2_model.parse(
        """
        fsig(1)
        sgen pzt m2.mech.z
        pd2 pd2 m1.p1.o eom.f 0 fsig.f
        """
    )
    pd2_sol = pd2_model.run(Xaxis(pd2_model.fsig.f, "log", f_min, f_max, steps))
    np.testing.assert_allclose(pd2_sol["pd2"], readout_tf)



"""Methods for saving outputs to files.

Currently just uses pickles until HDF is working.
"""

import importlib
import json
import os.path
import pickle
from collections import defaultdict

import numpy as np

from finesse.env import warn

# NOTE ssl: h5py is not available on OSX for ARM64 as of 2022-01-20; see #448.
try:
    import h5py
except ImportError:
    HAS_HDF5 = False
else:
    HAS_HDF5 = True

HDF_EXTENSIONS = [".h5", ".hdf5", ".hdf"]
PICKLE_EXTENSIONS = [".pkl", ".pickle"]
EXTENSIONS = HDF_EXTENSIONS + PICKLE_EXTENSIONS
FORMAT_OPTIONS = ["hdf", "pickle"]


def type_to_json(_type):
    if not isinstance(_type, type):
        raise TypeError(f"Not a type: {_type}")
    if hasattr(_type, "__module__"):
        return f"{_type.__module__}.{_type.__name__}"
    else:
        return _type.__name__


def type_from_json(data):
    # Grab the class name and figure out what type of object
    # to instantiate
    cls_name, *module = data.rsplit(".", maxsplit=1)[::-1]
    if len(module) == 0:  # builtin
        return getattr(__builtins__, cls_name)
    else:
        module = importlib.import_module(module[0])
        return getattr(module, cls_name)


def np_dtype_to_json(dtype):
    """Converts a numpy dtype into a json string format."""
    if not isinstance(dtype, np.dtype):
        raise TypeError(f"Not a dtype {dtype}")
    return json.dumps(tuple((a, b, tuple(c)) for a, b, *c in dtype.descr))


def np_dtype_from_json(data):
    data = json.loads(data)

    def process(a, b, c):
        a = a.strip()
        f = (
            lambda x: ()
            if len(x) == 0
            else tuple(np.atleast_1d(np.squeeze(x)).tolist())
        )
        c = f(c)
        if len(c) == 0:
            return (
                a,
                b,
            )
        else:
            return (a, b, c)

    # multiple checks needed to remove extra information stored during
    # the storage stage. If no name or shapes are given then we have
    # to go back to a single dtype string, rather than a tuple otherwise
    # you get a record array and an auto generated name for the column
    descr = [process(a, b, c) for a, b, *c in data]
    if len(descr) == 1 and len(descr[0]) == 2 and len(descr[0][0]) == 0:
        descr = descr[0][1]
    return np.dtype(descr)


def dict_to_json(d):
    return json.dumps(d)


def object_to_hdf(obj):
    pkl = pickle.dumps(obj, protocol=5)
    return np.void(pkl)


def create_object_dataset(grp, key, obj):
    d = grp.create_dataset(key, data=object_to_hdf(obj))
    d.attrs["type"] = type_to_json(type(obj))


def str_array_to_hdf(s):
    return np.bytes_(s)


def to_generic_hdf(obj, grp):
    """Used to convert a generic Python class into a HDF group.

    Scalar values (as determined by np.isscalar) are added as HDF group attributes. This
    includes class attributes that are int, float, strings, etc. Attributes that are
    representable with a numpy array are written as datasets within the group. Any
    attribute that does not fit in the above will be pickled and added as a byte stream
    dataset.
    """
    if not hasattr(obj, "__dict__"):
        warn(f"Nothing to write for {repr(obj)}")
        return

    for key, value in obj.__dict__.items():
        try:
            arr = np.asarray(value)
            if np.isscalar(value):
                # Write scalars as attributes to the group
                grp.attrs[key] = value
            elif arr.dtype.char == "U":  # array of strings
                grp.create_dataset(key, data=str_array_to_hdf(value))
            elif arr.dtype.char == "O":  # Object
                create_object_dataset(grp, key, value)
            else:  # Otherwise just try and dump the numpy array
                grp.create_dataset(key, data=value)
        except Exception as ex:
            raise Exception(f"Error writing {key}:{value} to HDF", ex)


def dump_solution_hdf(sol, filename):
    if not HAS_HDF5:
        raise RuntimeError("h5py not available; cannot dump to HDF. See #448.")

    def _dump(grp, obj):
        grp.attrs["__solution__.name"] = obj.name
        grp.attrs["__solution__.type"] = type_to_json(type(obj))
        dumpfunc = dump_mapping.get(type(obj), to_generic_hdf)
        grp.attrs["__solution__.dumper"] = f"{dumpfunc.__module__}.{dumpfunc.__name__}"
        dumpfunc(obj, grp)

    data = defaultdict(list)
    data[sol.name].append(sol)
    for _ in sol.get_all_children():
        data[_.get_path()].append(_)
    with h5py.File(filename, "w") as f:
        # Every solution object has its own group
        for path, sols in data.items():
            grp = f.create_group(path)
            if len(sols) > 1:
                # Then we have multiple solutions with the same name at this level
                for i, s in enumerate(sols):
                    g = grp.create_group(str(i))
                    _dump(g, s)
            else:
                _dump(grp, sols[0])


def load_solution_hdf(filename):
    if not HAS_HDF5:
        raise RuntimeError("h5py not available; cannot dump to HDF. See #448.")

    def _load_group(data, parent=None):
        groups = list(filter(lambda x: isinstance(x[1], h5py.Group), data.items()))
        datasets = list(filter(lambda x: isinstance(x[1], h5py.Dataset), data.items()))
        new_sol = None

        if "__solution__.type" in data.attrs:
            sol_type = data.attrs["__solution__.type"]
            sol_name = data.attrs["__solution__.name"]
            # Grab the class name and figure out what type of object
            # to instantiate
            cls_name, *module = sol_type.rsplit(".", maxsplit=1)[::-1]
            if len(module) == 0:  # builtin
                stype = getattr(__builtins__, cls_name)
            else:
                module = importlib.import_module(module[0])
                stype = getattr(module, cls_name)
            # Try and make a new object
            if stype in load_mapping:
                # Need to do more complicated mapping
                # this function should do all the attr
                # and dataset reading from the group
                # into the new object
                new_sol = load_mapping[stype](data, parent)
                if new_sol not in parent.children:
                    parent.add(new_sol)
                if type(new_sol) is not stype:
                    raise TypeError(
                        f"Excepted {load_mapping[stype]} to return an object of type {stype} not {new_sol}"
                    )
            else:
                # assume the most basic of interface for BaseSolution
                new_sol = stype(sol_name, parent=parent)
                # Set any scalar like attrs
                for attr in data.attrs:
                    if not attr.startswith("__solution__"):  # Ignore any metadata
                        setattr(new_sol, attr, data.attrs[attr])
                for name, ds in datasets:
                    setattr(new_sol, name, ds[()])

            parent = new_sol  # new parent to use as creating a new solution
        # Now load any other sub-groups which will be other solutions/groups
        for _, value in groups:
            _load_group(value, parent=parent)
        if new_sol:
            return new_sol

    with h5py.File(filename, "r") as f:
        groups = list(filter(lambda x: isinstance(x[1], h5py.Group), f.items()))
        datasets = list(filter(lambda x: isinstance(x[1], h5py.Dataset), f.items()))
        if len(datasets):
            warn("Found datasets in root which wasn't expected")
        if len(groups) != 1:
            raise Exception("Unexpected number of groups in root")
        return _load_group(groups[0][1], None)


def save(obj, filename, format=None):
    """Saves a Finesse solution object to a file. Two options are available: HDF5 and
    Pickle.

    Parameters
    ----------
    obj : Solution
        Solution object generated by a Finesse simulation
    filename : str
        A path and filename to save the output. If the path does not exist it will
        be created.
    format : str, optional
        For HDF files use one of ".h5", ".hdf5", ".hdf" or
        pickle files use one of ".pkl", ".pickle".
        If `None` then the extension of the `filename` is used.
    """
    if format is None:
        _, ext = os.path.splitext(filename)
        if ext in HDF_EXTENSIONS:
            format = "hdf"
        elif ext in PICKLE_EXTENSIONS:
            format = "pickle"
        elif ext is None:
            raise ValueError(
                "No file extension was provided, could not automatically choose format to use."
            )
        else:
            raise ValueError(
                f"{ext} not a supported file extension. Valid extensions are {EXTENSIONS}"
            )
    elif format not in FORMAT_OPTIONS:
        raise ValueError(f"format options are {FORMAT_OPTIONS}")

    a, _ = os.path.split(filename)
    if not os.path.exists(a) and len(a) > 0:
        os.mkdir(a)

    if format == "pickle":
        pickle.dump(obj, open(filename, "wb"), protocol=5)
    elif format == "hdf":
        dump_solution_hdf(obj, filename)


def load(filename, format=None):
    if format is None:
        _, ext = os.path.splitext(filename)
        if ext in HDF_EXTENSIONS:
            format = "hdf"
        elif ext in PICKLE_EXTENSIONS:
            format = "pickle"
        elif ext is None:
            raise ValueError(
                "No file extension was provided, could not automatically choose format to use."
            )
        else:
            raise ValueError(
                f"{ext} not a supported file extension. Valid extensions are {EXTENSIONS}"
            )
    elif format not in FORMAT_OPTIONS:
        raise ValueError(f"format options are {FORMAT_OPTIONS}")

    if format == "pickle":
        return pickle.load(open(filename, "rb"))
    elif format == "hdf":
        return load_solution_hdf(filename)


# object type to HDF data map
dump_mapping = {}
# HDF data to object type map
load_mapping = {}
